# Copyright 2015 Cloudera Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
from pathlib import Path
from typing import Generator

import pytest

import ibis
from ibis.backends.tests.base import BackendTest, RoundHalfToEven

PG_USER = os.environ.get(
    'IBIS_TEST_POSTGRES_USER', os.environ.get('PGUSER', 'postgres')
)
PG_PASS = os.environ.get(
    'IBIS_TEST_POSTGRES_PASSWORD', os.environ.get('PGPASSWORD', 'postgres')
)
PG_HOST = os.environ.get(
    'IBIS_TEST_POSTGRES_HOST', os.environ.get('PGHOST', 'localhost')
)
PG_PORT = os.environ.get(
    'IBIS_TEST_POSTGRES_PORT', os.environ.get('PGPORT', 5432)
)
IBIS_TEST_POSTGRES_DB = os.environ.get(
    'IBIS_TEST_POSTGRES_DATABASE', os.environ.get('PGDATABASE', 'ibis_testing')
)


class TestConf(BackendTest, RoundHalfToEven):
    # postgres rounds half to even for double precision and half away from zero
    # for numeric and decimal

    returned_timestamp_unit = 's'

    @staticmethod
    def connect(data_directory: Path):
        user = os.environ.get(
            'IBIS_TEST_POSTGRES_USER', os.environ.get('PGUSER', 'postgres')
        )
        password = os.environ.get(
            'IBIS_TEST_POSTGRES_PASSWORD', os.environ.get('PGPASS', 'postgres')
        )
        host = os.environ.get(
            'IBIS_TEST_POSTGRES_HOST', os.environ.get('PGHOST', 'localhost')
        )
        port = os.environ.get(
            'IBIS_TEST_POSTGRES_PORT', os.environ.get('PGPORT', '5432')
        )
        database = os.environ.get(
            'IBIS_TEST_POSTGRES_DATABASE',
            os.environ.get('PGDATABASE', 'ibis_testing'),
        )
        return ibis.postgres.connect(
            host=host,
            port=port,
            user=user,
            password=password,
            database=database,
        )


def _random_identifier(suffix):
    return f'__ibis_test_{suffix}_{ibis.util.guid()}'


@pytest.fixture(scope='session')
def con():
    return ibis.postgres.connect(
        host=PG_HOST,
        user=PG_USER,
        password=PG_PASS,
        database=IBIS_TEST_POSTGRES_DB,
        port=PG_PORT,
    )


@pytest.fixture(scope='module')
def db(con):
    return con.database()


@pytest.fixture(scope='module')
def alltypes(db):
    return db.functional_alltypes


@pytest.fixture(scope='module')
def geotable(con):
    return con.table('geo')


@pytest.fixture(scope='module')
def df(alltypes):
    return alltypes.execute()


@pytest.fixture(scope='module')
def gdf(geotable):
    return geotable.execute()


@pytest.fixture(scope='module')
def at(alltypes):
    return alltypes.op().sqla_table


@pytest.fixture(scope='module')
def intervals(con):
    return con.table("intervals")


@pytest.fixture
def translate():
    from ibis.backends.postgres import Backend

    context = Backend.compiler.make_context()
    return lambda expr: (
        Backend.compiler.translator_class(expr, context).get_result()
    )


@pytest.fixture
def temp_table(con) -> Generator[str, None, None]:
    """
    Return a temporary table name.

    Parameters
    ----------
    con : ibis.postgres.PostgreSQLClient

    Yields
    ------
    name : string
        Random table name for a temporary usage.
    """
    name = _random_identifier('table')
    try:
        yield name
    finally:
        con.drop_table(name, force=True)



from pathlib import Path
from typing import Any

import dask.dataframe as dd
import pandas as pd
import pandas.testing as tm
import pytest

import ibis
from ibis.backends.pandas.tests.conftest import TestConf as PandasTest

NPARTITIONS = 2


@pytest.fixture(scope="module")
def npartitions():
    return NPARTITIONS


class TestConf(PandasTest):
    @staticmethod
    def connect(data_directory: Path):
        # Note - we use `dd.from_pandas(pd.read_csv(...))` instead of
        # `dd.read_csv` due to https://github.com/dask/dask/issues/6970

        return ibis.dask.connect(
            {
                'functional_alltypes': dd.from_pandas(
                    pd.read_csv(
                        str(data_directory / 'functional_alltypes.csv'),
                        index_col=None,
                        dtype={'bool_col': bool, 'string_col': str},
                        parse_dates=['timestamp_col'],
                        encoding='utf-8',
                    ),
                    npartitions=NPARTITIONS,
                ),
                'batting': dd.from_pandas(
                    pd.read_csv(str(data_directory / 'batting.csv')),
                    npartitions=NPARTITIONS,
                ),
                'awards_players': dd.from_pandas(
                    pd.read_csv(str(data_directory / 'awards_players.csv')),
                    npartitions=NPARTITIONS,
                ),
            }
        )

    @classmethod
    def assert_series_equal(
        cls, left: pd.DataFrame, right: pd.DataFrame, *args: Any, **kwargs: Any
    ) -> None:
        kwargs.setdefault('check_dtype', cls.check_dtype)
        kwargs.setdefault('check_names', cls.check_names)
        left = left.reset_index(drop=True)
        right = right.reset_index(drop=True)
        tm.assert_series_equal(left, right, *args, **kwargs)


@pytest.fixture
def dataframe(npartitions):
    return dd.from_pandas(
        pd.DataFrame(
            {
                'plain_int64': list(range(1, 4)),
                'plain_strings': list('abc'),
                'dup_strings': list('dad'),
            }
        ),
        npartitions=npartitions,
    )


@pytest.fixture
def core_client(dataframe):
    return ibis.dask.connect({'df': dataframe})


@pytest.fixture
def ibis_table(core_client):
    return core_client.table('df')



import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.expr.types as ir
import ibis.util as util

from . import aggregate, binary_infix, case, helpers, string, timestamp, window
from .literal import literal, null_literal


def fixed_arity(func_name, arity):
    def formatter(translator, expr):
        op = expr.op()
        if arity != len(op.args):
            raise com.IbisError('incorrect number of args')
        return helpers.format_call(translator, func_name, *op.args)

    return formatter


def unary(func_name):
    return fixed_arity(func_name, 1)


def not_null(translator, expr):
    formatted_arg = translator.translate(expr.op().args[0])
    return f'{formatted_arg} IS NOT NULL'


def is_null(translator, expr):
    formatted_arg = translator.translate(expr.op().args[0])
    return f'{formatted_arg} IS NULL'


def not_(translator, expr):
    (arg,) = expr.op().args
    formatted_arg = translator.translate(arg)
    if helpers.needs_parens(arg):
        formatted_arg = helpers.parenthesize(formatted_arg)
    return f'NOT {formatted_arg}'


def negate(translator, expr):
    arg = expr.op().args[0]
    formatted_arg = translator.translate(arg)
    if isinstance(expr, ir.BooleanValue):
        return not_(translator, expr)
    else:
        if helpers.needs_parens(arg):
            formatted_arg = helpers.parenthesize(formatted_arg)
        return f'-{formatted_arg}'


def ifnull_workaround(translator, expr):
    op = expr.op()
    a, b = op.args

    # work around per #345, #360
    if isinstance(a, ir.DecimalValue) and isinstance(b, ir.IntegerValue):
        b = b.cast(a.type())

    return helpers.format_call(translator, 'isnull', a, b)


def sign(translator, expr):
    (arg,) = expr.op().args
    translated_arg = translator.translate(arg)
    translated_type = helpers.type_to_sql_string(expr.type())
    if expr.type() != dt.float:
        return f'CAST(sign({translated_arg}) AS {translated_type})'
    return f'sign({translated_arg})'


def hashbytes(translator, expr):
    op = expr.op()
    arg, how = op.args

    arg_formatted = translator.translate(arg)

    if how == 'md5':
        return f'md5({arg_formatted})'
    elif how == 'sha1':
        return f'sha1({arg_formatted})'
    elif how == 'sha256':
        return f'sha256({arg_formatted})'
    elif how == 'sha512':
        return f'sha512({arg_formatted})'
    else:
        raise NotImplementedError(how)


def log(translator, expr):
    op = expr.op()
    arg, base = op.args
    arg_formatted = translator.translate(arg)

    if base is None:
        return f'ln({arg_formatted})'

    base_formatted = translator.translate(base)
    return f'log({base_formatted}, {arg_formatted})'


def value_list(translator, expr):
    op = expr.op()
    formatted = [translator.translate(x) for x in op.values]
    return helpers.parenthesize(', '.join(formatted))


def cast(translator, expr):
    op = expr.op()
    arg, target_type = op.args
    arg_formatted = translator.translate(arg)

    if isinstance(arg, ir.CategoryValue) and target_type == dt.int32:
        return arg_formatted
    if isinstance(arg, ir.TemporalValue) and target_type == dt.int64:
        return f'1000000 * unix_timestamp({arg_formatted})'
    else:
        sql_type = helpers.type_to_sql_string(target_type)
        return f'CAST({arg_formatted} AS {sql_type})'


def varargs(func_name):
    def varargs_formatter(translator, expr):
        op = expr.op()
        return helpers.format_call(translator, func_name, *op.arg)

    return varargs_formatter


def between(translator, expr):
    op = expr.op()
    comp, lower, upper = (translator.translate(x) for x in op.args)
    return f'{comp} BETWEEN {lower} AND {upper}'


def table_array_view(translator, expr):
    ctx = translator.context
    table = expr.op().table
    query = ctx.get_compiled_expr(table)
    return f'(\n{util.indent(query, ctx.indent)}\n)'


def table_column(translator, expr):
    op = expr.op()
    field_name = op.name
    quoted_name = helpers.quote_identifier(field_name, force=True)

    table = op.table
    ctx = translator.context

    # If the column does not originate from the table set in the current SELECT
    # context, we should format as a subquery
    if translator.permit_subquery and ctx.is_foreign_expr(table):
        proj_expr = table.projection([field_name]).to_array()
        return table_array_view(translator, proj_expr)

    if ctx.need_aliases():
        alias = ctx.get_ref(table)
        if alias is not None:
            quoted_name = f'{alias}.{quoted_name}'

    return quoted_name


def exists_subquery(translator, expr):
    op = expr.op()
    ctx = translator.context

    dummy = ir.literal(1).name(ir.unnamed)

    filtered = op.foreign_table.filter(op.predicates)
    expr = filtered.projection([dummy])

    subquery = ctx.get_compiled_expr(expr)

    if isinstance(op, ops.ExistsSubquery):
        key = 'EXISTS'
    elif isinstance(op, ops.NotExistsSubquery):
        key = 'NOT EXISTS'
    else:
        raise NotImplementedError

    return f'{key} (\n{util.indent(subquery, ctx.indent)}\n)'


# XXX this is not added to operation_registry, but looks like impala is
# using it in the tests, and it works, even if it's not imported anywhere
def round(translator, expr):
    op = expr.op()
    arg, digits = op.args

    arg_formatted = translator.translate(arg)

    if digits is not None:
        digits_formatted = translator.translate(digits)
        return f'round({arg_formatted}, {digits_formatted})'
    return f'round({arg_formatted})'


# XXX this is not added to operation_registry, but looks like impala is
# using it in the tests, and it works, even if it's not imported anywhere
def hash(translator, expr):
    op = expr.op()
    arg, how = op.args

    arg_formatted = translator.translate(arg)

    if how == 'fnv':
        return f'fnv_hash({arg_formatted})'
    else:
        raise NotImplementedError(how)


binary_infix_ops = {
    # Binary operations
    ops.Add: binary_infix.binary_infix_op('+'),
    ops.Subtract: binary_infix.binary_infix_op('-'),
    ops.Multiply: binary_infix.binary_infix_op('*'),
    ops.Divide: binary_infix.binary_infix_op('/'),
    ops.Power: fixed_arity('pow', 2),
    ops.Modulus: binary_infix.binary_infix_op('%'),
    # Comparisons
    ops.Equals: binary_infix.binary_infix_op('='),
    ops.NotEquals: binary_infix.binary_infix_op('!='),
    ops.GreaterEqual: binary_infix.binary_infix_op('>='),
    ops.Greater: binary_infix.binary_infix_op('>'),
    ops.LessEqual: binary_infix.binary_infix_op('<='),
    ops.Less: binary_infix.binary_infix_op('<'),
    ops.IdenticalTo: binary_infix.identical_to,
    # Boolean comparisons
    ops.And: binary_infix.binary_infix_op('AND'),
    ops.Or: binary_infix.binary_infix_op('OR'),
    ops.Xor: binary_infix.xor,
}


operation_registry = {
    # Unary operations
    ops.NotNull: not_null,
    ops.IsNull: is_null,
    ops.Negate: negate,
    ops.Not: not_,
    ops.IsNan: unary('is_nan'),
    ops.IsInf: unary('is_inf'),
    ops.IfNull: ifnull_workaround,
    ops.NullIf: fixed_arity('nullif', 2),
    ops.ZeroIfNull: unary('zeroifnull'),
    ops.NullIfZero: unary('nullifzero'),
    ops.Abs: unary('abs'),
    ops.BaseConvert: fixed_arity('conv', 3),
    ops.Ceil: unary('ceil'),
    ops.Floor: unary('floor'),
    ops.Exp: unary('exp'),
    ops.Round: round,
    ops.Sign: sign,
    ops.Sqrt: unary('sqrt'),
    ops.Hash: hash,
    ops.HashBytes: hashbytes,
    ops.Log: log,
    ops.Ln: unary('ln'),
    ops.Log2: unary('log2'),
    ops.Log10: unary('log10'),
    ops.DecimalPrecision: unary('precision'),
    ops.DecimalScale: unary('scale'),
    # Unary aggregates
    ops.CMSMedian: aggregate.reduction('appx_median'),
    ops.HLLCardinality: aggregate.reduction('ndv'),
    ops.Mean: aggregate.reduction('avg'),
    ops.Sum: aggregate.reduction('sum'),
    ops.Max: aggregate.reduction('max'),
    ops.Min: aggregate.reduction('min'),
    ops.StandardDev: aggregate.variance_like('stddev'),
    ops.Variance: aggregate.variance_like('var'),
    ops.GroupConcat: aggregate.reduction('group_concat'),
    ops.Count: aggregate.reduction('count'),
    ops.CountDistinct: aggregate.count_distinct,
    # string operations
    ops.StringLength: unary('length'),
    ops.StringAscii: unary('ascii'),
    ops.Lowercase: unary('lower'),
    ops.Uppercase: unary('upper'),
    ops.Reverse: unary('reverse'),
    ops.Strip: unary('trim'),
    ops.LStrip: unary('ltrim'),
    ops.RStrip: unary('rtrim'),
    ops.Capitalize: unary('initcap'),
    ops.Substring: string.substring,
    ops.StrRight: fixed_arity('strright', 2),
    ops.Repeat: fixed_arity('repeat', 2),
    ops.StringFind: string.string_find,
    ops.Translate: fixed_arity('translate', 3),
    ops.FindInSet: string.find_in_set,
    ops.LPad: fixed_arity('lpad', 3),
    ops.RPad: fixed_arity('rpad', 3),
    ops.StringJoin: string.string_join,
    ops.StringSQLLike: string.string_like,
    ops.RegexSearch: fixed_arity('regexp_like', 2),
    ops.RegexExtract: fixed_arity('regexp_extract', 3),
    ops.RegexReplace: fixed_arity('regexp_replace', 3),
    ops.ParseURL: string.parse_url,
    ops.StartsWith: string.startswith,
    ops.EndsWith: string.endswith,
    # Timestamp operations
    ops.Date: unary('to_date'),
    ops.TimestampNow: lambda *args: 'now()',
    ops.ExtractYear: timestamp.extract_field('year'),
    ops.ExtractMonth: timestamp.extract_field('month'),
    ops.ExtractDay: timestamp.extract_field('day'),
    ops.ExtractQuarter: timestamp.extract_field('quarter'),
    ops.ExtractEpochSeconds: timestamp.extract_epoch_seconds,
    ops.ExtractWeekOfYear: fixed_arity('weekofyear', 1),
    ops.ExtractHour: timestamp.extract_field('hour'),
    ops.ExtractMinute: timestamp.extract_field('minute'),
    ops.ExtractSecond: timestamp.extract_field('second'),
    ops.ExtractMillisecond: timestamp.extract_field('millisecond'),
    ops.TimestampTruncate: timestamp.truncate,
    ops.DateTruncate: timestamp.truncate,
    ops.IntervalFromInteger: timestamp.interval_from_integer,
    # Other operations
    ops.E: lambda *args: 'e()',
    ops.Literal: literal,
    ops.NullLiteral: null_literal,
    ops.ValueList: value_list,
    ops.Cast: cast,
    ops.Coalesce: varargs('coalesce'),
    ops.Greatest: varargs('greatest'),
    ops.Least: varargs('least'),
    ops.Where: fixed_arity('if', 3),
    ops.Between: between,
    ops.Contains: binary_infix.binary_infix_op('IN'),
    ops.NotContains: binary_infix.binary_infix_op('NOT IN'),
    ops.SimpleCase: case.simple_case,
    ops.SearchedCase: case.searched_case,
    ops.TableColumn: table_column,
    ops.TableArrayView: table_array_view,
    ops.DateAdd: timestamp.timestamp_op('date_add'),
    ops.DateSub: timestamp.timestamp_op('date_sub'),
    ops.DateDiff: timestamp.timestamp_op('datediff'),
    ops.TimestampAdd: timestamp.timestamp_op('date_add'),
    ops.TimestampSub: timestamp.timestamp_op('date_sub'),
    ops.TimestampDiff: timestamp.timestamp_diff,
    ops.TimestampFromUNIX: timestamp.timestamp_from_unix,
    ops.ExistsSubquery: exists_subquery,
    ops.NotExistsSubquery: exists_subquery,
    # RowNumber, and rank functions starts with 0 in Ibis-land
    ops.RowNumber: lambda *args: 'row_number()',
    ops.DenseRank: lambda *args: 'dense_rank()',
    ops.MinRank: lambda *args: 'rank()',
    ops.PercentRank: lambda *args: 'percent_rank()',
    ops.FirstValue: unary('first_value'),
    ops.LastValue: unary('last_value'),
    ops.NthValue: window.nth_value,
    ops.Lag: window.shift_like('lag'),
    ops.Lead: window.shift_like('lead'),
    ops.WindowOp: window.window,
    ops.NTile: window.ntile,
    ops.DayOfWeekIndex: timestamp.day_of_week_index,
    ops.DayOfWeekName: timestamp.day_of_week_name,
    **binary_infix_ops,
}



import pandas as pd
import pandas.testing as tm
import pytest

import ibis
import ibis.expr.datatypes as dt
from ibis.util import guid

SQLALCHEMY_BACKENDS = ['sqlite', 'postgres', 'mysql']


@pytest.fixture
def new_schema():
    return ibis.schema([('a', 'string'), ('b', 'bool'), ('c', 'int32')])


@pytest.mark.only_on_backends(
    SQLALCHEMY_BACKENDS,
    reason="run only if backend is SQLAlchemy based",
)
@pytest.mark.xfail_unsupported
def test_load_data_sqlalchemy(backend, con, temp_table):
    sch = ibis.schema(
        [
            ('first_name', 'string'),
            ('last_name', 'string'),
            ('department_name', 'string'),
            ('salary', 'float64'),
        ]
    )

    df = pd.DataFrame(
        {
            'first_name': ['A', 'B', 'C'],
            'last_name': ['D', 'E', 'F'],
            'department_name': ['AA', 'BB', 'CC'],
            'salary': [100.0, 200.0, 300.0],
        }
    )
    con.create_table(temp_table, schema=sch)
    con.load_data(temp_table, df, if_exists='append')
    result = con.table(temp_table).execute()

    backend.assert_frame_equal(df, result)


@pytest.mark.parametrize(
    ('expr_fn', 'expected'),
    [
        (lambda t: t.string_col, [('string_col', dt.String)]),
        (
            lambda t: t[t.string_col, t.bigint_col],
            [('string_col', dt.String), ('bigint_col', dt.Int64)],
        ),
    ],
)
def test_query_schema(backend, con, alltypes, expr_fn, expected):
    if not hasattr(con, 'compiler'):
        pytest.skip(
            '{} backend has no `compiler` attribute'.format(
                type(backend).__name__
            )
        )

    expr = expr_fn(alltypes)

    # we might need a public API for it
    ast = con.compiler.to_ast(expr, backend.make_context())
    schema = con.ast_schema(ast)

    # clickhouse columns has been defined as non-nullable
    # whereas other backends don't support non-nullable columns yet
    expected = ibis.schema(
        [
            (name, dtype(nullable=schema[name].nullable))
            for name, dtype in expected
        ]
    )
    assert schema.equals(expected)


@pytest.mark.parametrize(
    'sql',
    [
        'select * from functional_alltypes limit 10',
        'select * from functional_alltypes \nlimit 10\n',
    ],
)
@pytest.mark.xfail_backends(('bigquery',))
@pytest.mark.xfail_unsupported
def test_sql(backend, con, sql):
    if not hasattr(con, 'sql') or not hasattr(con, '_get_schema_using_query'):
        pytest.skip(f'Backend {backend} does not support sql method')

    # execute the expression using SQL query
    con.sql(sql).execute()


# test table


@pytest.mark.xfail_unsupported
@pytest.mark.xfail_backends(['pandas', 'dask'])
def test_create_table_from_schema(con, backend, new_schema, temp_table):
    # xfailing pandas and dask: #3020
    con.create_table(temp_table, schema=new_schema)

    t = con.table(temp_table)

    for k, i_type in t.schema().items():
        assert new_schema[k] == i_type


@pytest.mark.xfail_unsupported
def test_rename_table(con, backend, temp_table, new_schema):
    if not hasattr(con, 'rename_table'):
        pytest.xfail('{} backend doesn\'t have rename_table method.')

    temp_table_original = f'{temp_table}_original'
    con.create_table(temp_table_original, schema=new_schema)

    t = con.table(temp_table_original)
    t.rename(temp_table)

    assert con.table(temp_table) is not None
    assert temp_table in con.list_tables()


@pytest.mark.xfail_unsupported
@pytest.mark.xfail_backends(['impala', 'pyspark', 'spark', 'pandas', 'dask'])
def test_nullable_input_output(con, backend, temp_table):
    # - Impala, PySpark and Spark non-nullable issues #2138 and #2137
    # xfailing pandas and dask: #3020
    sch = ibis.schema(
        [
            ('foo', 'int64'),
            ('bar', ibis.expr.datatypes.int64(nullable=False)),
            ('baz', 'boolean'),
        ]
    )

    con.create_table(temp_table, schema=sch)

    t = con.table(temp_table)

    assert t.schema().types[0].nullable
    assert not t.schema().types[1].nullable
    assert t.schema().types[2].nullable


# view tests


@pytest.mark.xfail_unsupported
@pytest.mark.xfail_backends(['pyspark', 'spark'])
def test_create_drop_view(con, backend, temp_view):
    # pyspark and spark skipt because table actually is a temporary view

    # setup
    table_name = 'functional_alltypes'
    expr = con.table(table_name).limit(1)

    # create a new view
    con.create_view(temp_view, expr)
    # check if the view was created
    assert temp_view in con.list_tables()

    t_expr = con.table(table_name)
    v_expr = con.table(temp_view)
    # check if the view and the table has the same fields
    assert set(t_expr.schema().names) == set(v_expr.schema().names)


@pytest.mark.only_on_backends(
    ['bigquery', 'clickhouse', 'impala', 'omniscidb', 'spark'],
    reason="run only if backend is sql-based",
)
def test_separate_database(con, alternate_current_database, current_data_db):
    # using alternate_current_database switches "con" current
    #  database to a temporary one until a test is over
    tmp_db = con.database(alternate_current_database)
    # verifying we can open another db which isn't equal to current
    db = con.database(current_data_db)
    assert db.name == current_data_db
    assert tmp_db.name == alternate_current_database


def _create_temp_table_with_schema(con, temp_table_name, schema, data=None):

    con.drop_table(temp_table_name, force=True)
    con.create_table(temp_table_name, schema=schema)
    temporary = con.table(temp_table_name)
    assert len(temporary.execute()) == 0

    if data is not None and isinstance(data, pd.DataFrame):
        con.load_data(temp_table_name, data, if_exists='append')
        assert len(temporary.execute()) == len(data.index)
        tm.assert_frame_equal(temporary.execute(), data)

    return temporary


@pytest.mark.only_on_backends(
    SQLALCHEMY_BACKENDS,
    reason="run only if backend is SQLAlchemy based",
)
def test_insert_no_overwrite_from_dataframe(
    con, test_employee_schema, test_employee_data_2
):

    temp_table = f'temp_to_table_{guid()}'
    temporary = _create_temp_table_with_schema(
        con,
        temp_table,
        test_employee_schema,
    )

    con.insert(temp_table, obj=test_employee_data_2, overwrite=False)
    assert len(temporary.execute()) == 3
    tm.assert_frame_equal(temporary.execute(), test_employee_data_2)


@pytest.mark.only_on_backends(
    SQLALCHEMY_BACKENDS,
    reason="run only if backend is SQLAlchemy based",
)
def test_insert_overwrite_from_dataframe(
    con, test_employee_schema, test_employee_data_1, test_employee_data_2
):

    temp_table = f'temp_to_table_{guid()}'
    temporary = _create_temp_table_with_schema(
        con,
        temp_table,
        test_employee_schema,
        data=test_employee_data_1,
    )

    con.insert(temp_table, obj=test_employee_data_2, overwrite=True)
    assert len(temporary.execute()) == 3
    tm.assert_frame_equal(temporary.execute(), test_employee_data_2)


@pytest.mark.only_on_backends(
    SQLALCHEMY_BACKENDS,
    reason="run only if backend is SQLAlchemy based",
)
def test_insert_no_overwite_from_expr(
    con, test_employee_schema, test_employee_data_2
):

    temp_table = f'temp_to_table_{guid()}'
    temporary = _create_temp_table_with_schema(
        con,
        temp_table,
        test_employee_schema,
    )

    from_table_name = f'temp_from_table_{guid()}'
    from_table = _create_temp_table_with_schema(
        con,
        from_table_name,
        test_employee_schema,
        data=test_employee_data_2,
    )

    con.insert(temp_table, obj=from_table, overwrite=False)
    assert len(temporary.execute()) == 3
    tm.assert_frame_equal(temporary.execute(), from_table.execute())


@pytest.mark.only_on_backends(
    SQLALCHEMY_BACKENDS,
    reason="run only if backend is SQLAlchemy based",
)
def test_insert_overwrite_from_expr(
    con, test_employee_schema, test_employee_data_1, test_employee_data_2
):

    temp_table = f'temp_to_table_{guid()}'
    temporary = _create_temp_table_with_schema(
        con,
        temp_table,
        test_employee_schema,
        data=test_employee_data_1,
    )

    from_table_name = f'temp_from_table_{guid()}'
    from_table = _create_temp_table_with_schema(
        con,
        from_table_name,
        test_employee_schema,
        data=test_employee_data_2,
    )

    con.insert(temp_table, obj=from_table, overwrite=True)
    assert len(temporary.execute()) == 3
    tm.assert_frame_equal(temporary.execute(), from_table.execute())


@pytest.mark.only_on_backends(
    SQLALCHEMY_BACKENDS,
    reason="run only if backend is SQLAlchemy based",
)
def test_list_databases(con):
    # Every backend has its own databases
    TEST_DATABASES = {
        'sqlite': ['main', 'base'],
        'postgres': ['postgres', 'ibis_testing'],
        'mysql': ['ibis_testing', 'information_schema'],
    }
    assert con.list_databases() == TEST_DATABASES[con.name]


@pytest.mark.only_on_backends(
    set(SQLALCHEMY_BACKENDS) - {'postgres'},
    reason="run only if backend is SQLAlchemy based, except postgres which "
    "which has schemas different than databases",
)
def test_list_schemas(con):
    with pytest.warns(FutureWarning):
        schemas = con.list_schemas()
    assert schemas == con.list_databases()


def test_verify(con, backend):
    if not hasattr(con, 'compiler'):
        pytest.skip(
            '{} backend has no `compiler` attribute'.format(
                type(backend).__name__
            )
        )
    expr = con.table('functional_alltypes').double_col.sum()

    with pytest.warns(FutureWarning):
        assert expr.verify()

    with pytest.warns(FutureWarning):
        assert backend.api.verify(expr)

    # There is no expression that can't be compiled to any backend
    # Testing `not verify()` only for an expression not supported in postgres
    if backend.api.name == 'postgres':
        expr = con.table('functional_alltypes').double_col.approx_median()
        with pytest.warns(FutureWarning):
            assert not expr.verify()

        with pytest.warns(FutureWarning):
            assert not backend.api.verify(expr)



"""Execution rules for ops.Where operations"""

import dask.dataframe as dd
import numpy as np

import ibis.expr.operations as ops
from ibis.backends.pandas.core import boolean_types, scalar_types
from ibis.backends.pandas.execution.generic import (
    execute_node_where_scalar_scalar_scalar,
    execute_node_where_series_series_series,
)

from ..dispatch import execute_node
from .util import TypeRegistrationDict, register_types_to_dispatcher

DASK_DISPATCH_TYPES: TypeRegistrationDict = {
    ops.Where: [
        (
            (dd.Series, dd.Series, dd.Series),
            execute_node_where_series_series_series,
        ),
        (
            (dd.Series, dd.Series, scalar_types),
            execute_node_where_series_series_series,
        ),
        (
            (
                boolean_types,
                dd.Series,
                dd.Series,
            ),
            execute_node_where_scalar_scalar_scalar,
        ),
    ]
}
register_types_to_dispatcher(execute_node, DASK_DISPATCH_TYPES)


def execute_node_where_series_scalar_scalar(op, cond, true, false, **kwargs):
    return dd.from_array(np.repeat(true, len(cond))).where(cond, other=false)


for scalar_type in scalar_types:
    execute_node.register(ops.Where, dd.Series, scalar_type, scalar_type)(
        execute_node_where_series_scalar_scalar
    )


@execute_node.register(ops.Where, boolean_types, dd.Series, scalar_types)
def execute_node_where_scalar_series_scalar(op, cond, true, false, **kwargs):
    if cond:
        return true
    else:
        # TODO double check this is the right way to do this
        out = dd.from_array(np.repeat(false, len(true)))
        out.index = true.index
        return out


@execute_node.register(ops.Where, boolean_types, scalar_types, dd.Series)
def execute_node_where_scalar_scalar_series(op, cond, true, false, **kwargs):
    return dd.from_array(np.repeat(true, len(false))) if cond else false



import functools
import inspect
import math

import regex as re

_SQLITE_UDF_REGISTRY = set()
_SQLITE_UDAF_REGISTRY = set()


def udf(f):
    """Create a SQLite scalar UDF from `f`

    Parameters
    ----------
    f
        A callable object

    Returns
    -------
    callable
        A callable object that returns ``None`` if any of its inputs are
        ``None``.
    """

    @functools.wraps(f)
    def wrapper(*args):
        if any(arg is None for arg in args):
            return None
        return f(*args)

    _SQLITE_UDF_REGISTRY.add(wrapper)
    return wrapper


def udaf(cls):
    """Register a UDAF class with any SQLite connection."""
    _SQLITE_UDAF_REGISTRY.add(cls)
    return cls


@udf
def _ibis_sqlite_reverse(string):
    return string[::-1]


@udf
def _ibis_sqlite_string_ascii(string):
    return ord(string[0])


@udf
def _ibis_sqlite_capitalize(string):
    return string.capitalize()


@udf
def _ibis_sqlite_translate(string, from_string, to_string):
    table = str.maketrans(from_string, to_string)
    return string.translate(table)


@udf
def _ibis_sqlite_regex_search(string, regex):
    """Return whether `regex` exists in `string`.

    Parameters
    ----------
    string : str
    regex : str

    Returns
    -------
    found : bool
    """
    return re.search(regex, string) is not None


@udf
def _ibis_sqlite_regex_replace(string, pattern, replacement):
    """Replace occurences of `pattern` in `string` with `replacement`.

    Parameters
    ----------
    string : str
    pattern : str
    replacement : str

    Returns
    -------
    result : str
    """
    return re.sub(pattern, replacement, string)


@udf
def _ibis_sqlite_regex_extract(string, pattern, index):
    """Extract match of regular expression `pattern` from `string` at `index`.

    Parameters
    ----------
    string : str
    pattern : str
    index : int

    Returns
    -------
    result : str or None
    """
    result = re.search(pattern, string)
    if result is not None and 0 <= index <= (result.lastindex or -1):
        return result.group(index)
    return None


@udf
def _ibis_sqlite_exp(arg):
    """Exponentiate `arg`.

    Parameters
    ----------
    arg : number
        Number to raise to `e`.

    Returns
    -------
    result : Optional[number]
        None If the input is None
    """
    return math.exp(arg)


@udf
def _ibis_sqlite_log(arg, base):
    if arg < 0 or base < 0:
        return None
    return math.log(arg, base)


@udf
def _ibis_sqlite_ln(arg):
    if arg < 0:
        return None
    return math.log(arg)


@udf
def _ibis_sqlite_log2(arg):
    return _ibis_sqlite_log(arg, 2)


@udf
def _ibis_sqlite_log10(arg):
    return _ibis_sqlite_log(arg, 10)


@udf
def _ibis_sqlite_floor(arg):
    return math.floor(arg)


@udf
def _ibis_sqlite_ceil(arg):
    return math.ceil(arg)


@udf
def _ibis_sqlite_sign(arg):
    if not arg:
        return 0
    return math.copysign(1, arg)


@udf
def _ibis_sqlite_floordiv(left, right):
    return left // right


@udf
def _ibis_sqlite_mod(left, right):
    return left % right


@udf
def _ibis_sqlite_power(arg, power):
    """Raise `arg` to the `power` power.

    Parameters
    ----------
    arg : number
        Number to raise to `power`.
    power : number
        Number to raise `arg` to.

    Returns
    -------
    result : Optional[number]
        None If either argument is None or we're trying to take a fractional
        power or a negative number
    """
    if arg < 0.0 and not power.is_integer():
        return None
    return arg ** power


@udf
def _ibis_sqlite_sqrt(arg):
    """Square root of `arg`.

    Parameters
    ----------
    arg : Optional[number]
        Number to take the square root of

    Returns
    -------
    result : Optional[number]
        None if `arg` is None or less than 0 otherwise the square root
    """
    return None if arg is None or arg < 0.0 else math.sqrt(arg)


class _ibis_sqlite_var:
    def __init__(self, offset):
        self.mean = 0.0
        self.sum_of_squares_of_differences = 0.0
        self.count = 0
        self.offset = offset

    def step(self, value):
        if value is not None:
            self.count += 1
            delta = value - self.mean
            self.mean += delta / self.count
            self.sum_of_squares_of_differences += delta * (value - self.mean)

    def finalize(self):
        count = self.count
        if count:
            return self.sum_of_squares_of_differences / (count - self.offset)
        return None


@udaf
class _ibis_sqlite_var_pop(_ibis_sqlite_var):
    def __init__(self):
        super().__init__(0)


@udaf
class _ibis_sqlite_var_samp(_ibis_sqlite_var):
    def __init__(self):
        super().__init__(1)


def _number_of_arguments(callable):
    signature = inspect.signature(callable)
    parameters = signature.parameters.values()
    kinds = [param.kind for param in parameters]
    valid_kinds = (
        inspect.Parameter.POSITIONAL_OR_KEYWORD,
        inspect.Parameter.POSITIONAL_ONLY,
    )
    if any(kind not in valid_kinds for kind in kinds) or any(
        param.default is not inspect.Parameter.empty for param in parameters
    ):
        raise TypeError(
            'Only positional arguments without defaults are supported in Ibis '
            'SQLite function registration'
        )
    return len(parameters)


def _register_function(func, con):
    """Register a Python callable with a SQLite connection `con`.

    Parameters
    ----------
    func : callable
    con : sqlalchemy.Connection
    """
    nargs = _number_of_arguments(func)
    con.connection.connection.create_function(func.__name__, nargs, func)


def _register_aggregate(agg, con):
    """Register a Python class that performs aggregation in SQLite.

    Parameters
    ----------
    agg : type
    con : sqlalchemy.Connection
    """
    nargs = _number_of_arguments(agg.step) - 1  # because self
    con.connection.connection.create_aggregate(agg.__name__, nargs, agg)


def register_all(con):
    """Register all udf and udaf with the connection.

    All udf and udaf are defined in this file with the `udf` and `udaf`
    decorators.

    Parameters
    ----------
    con : sqlalchemy.Connection
    """
    for func in _SQLITE_UDF_REGISTRY:
        con.run_callable(functools.partial(_register_function, func))

    for agg in _SQLITE_UDAF_REGISTRY:
        con.run_callable(functools.partial(_register_aggregate, agg))



"""APIs for creating user-defined element-wise, reduction and analytic
functions.
"""


import collections
import functools
import itertools
from typing import Tuple

import numpy as np
import pandas as pd
from pandas.core.groupby import SeriesGroupBy

import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.udf.vectorized
from ibis.backends.base import BaseBackend

from .aggcontext import Transform
from .core import date_types, time_types, timedelta_types, timestamp_types
from .dispatch import execute_node, pre_execute


@functools.singledispatch
def rule_to_python_type(datatype):
    """Convert an ibis :class:`~ibis.expr.datatypes.DataType` into a pandas
    backend friendly ``multipledispatch`` signature.

    Parameters
    ----------
    rule : DataType
        The :class:`~ibis.expr.datatypes.DataType` subclass to map to a pandas
        friendly type.

    Returns
    -------
    Union[Type[U], Tuple[Type[T], ...]]
        A pandas-backend-friendly signature
    """
    raise NotImplementedError(
        "Don't know how to convert type {} into a native Python type".format(
            type(datatype)
        )
    )


def create_gens_from_args_groupby(args: Tuple[SeriesGroupBy]):
    """Create generators for each args for groupby udaf.

    Returns a generator that outputs each group.

    Parameters
    ----------
    args : Tuple[SeriesGroupBy...]

    Returns
    -------
    Tuple[Generator]
    """
    iters = ((data for _, data in arg) for arg in args)
    return iters


@rule_to_python_type.register(dt.Array)
def array_rule(rule):
    return (list,)


@rule_to_python_type.register(dt.Map)
def map_rule(rule):
    return (dict,)


@rule_to_python_type.register(dt.Struct)
def struct_rule(rule):
    return (collections.OrderedDict,)


@rule_to_python_type.register(dt.String)
def string_rule(rule):
    return (str,)


@rule_to_python_type.register(dt.Integer)
def int_rule(rule):
    return int, np.integer


@rule_to_python_type.register(dt.Floating)
def float_rule(rule):
    return float, np.floating


@rule_to_python_type.register(dt.Boolean)
def bool_rule(rule):
    return bool, np.bool_


@rule_to_python_type.register(dt.Interval)
def interval_rule(rule):
    return timedelta_types


@rule_to_python_type.register(dt.Date)
def date_rule(rule):
    return date_types


@rule_to_python_type.register(dt.Timestamp)
def timestamp_rule(rule):
    return timestamp_types


@rule_to_python_type.register(dt.Time)
def time_rule(rule):
    return time_types


def nullable(datatype):
    """Return the signature of a scalar value that is allowed to be NULL (in
    SQL parlance).

    Parameters
    ----------
    datatype : ibis.expr.datatypes.DataType

    Returns
    -------
    Tuple[Type]
    """
    return (type(None),) if datatype.nullable else ()


class udf:
    @staticmethod
    def elementwise(input_type, output_type):
        """Alias for ibis.udf.vectorized.elementwise."""

        return ibis.udf.vectorized.elementwise(input_type, output_type)

    @staticmethod
    def reduction(input_type, output_type):
        """Alias for ibis.udf.vectorized.reduction."""
        return ibis.udf.vectorized.reduction(input_type, output_type)

    @staticmethod
    def analytic(input_type, output_type):
        """Alias for ibis.udf.vectorized.analytic."""
        return ibis.udf.vectorized.analytic(input_type, output_type)


@pre_execute.register(ops.ElementWiseVectorizedUDF)
@pre_execute.register(ops.ElementWiseVectorizedUDF, BaseBackend)
def pre_execute_elementwise_udf(op, *clients, scope=None, **kwargs):
    """Register execution rules for elementwise UDFs."""
    input_type = op.input_type

    # definitions

    # Define an execution rule for elementwise operations on a
    # grouped Series
    nargs = len(input_type)

    @execute_node.register(
        ops.ElementWiseVectorizedUDF, *(itertools.repeat(SeriesGroupBy, nargs))
    )
    def execute_udf_node_groupby(op, *args, **kwargs):
        func = op.func

        groupers = [
            grouper
            for grouper in (getattr(arg, 'grouper', None) for arg in args)
            if grouper is not None
        ]

        # all grouping keys must be identical
        assert all(groupers[0] == grouper for grouper in groupers[1:])

        # we're performing a scalar operation on grouped column, so
        # perform the operation directly on the underlying Series
        # and regroup after it's finished
        args = [getattr(arg, 'obj', arg) for arg in args]
        groupings = groupers[0].groupings
        return func(*args).groupby(groupings)

    # Define an execution rule for a simple elementwise Series
    # function
    @execute_node.register(
        ops.ElementWiseVectorizedUDF, *(itertools.repeat(pd.Series, nargs))
    )
    @execute_node.register(
        ops.ElementWiseVectorizedUDF, *(itertools.repeat(object, nargs))
    )
    def execute_udf_node(op, *args, **kwargs):
        # We have rewritten op.func to be a closure enclosing
        # the kwargs, and therefore, we do not need to pass
        # kwargs here. This is true for all udf execution in this
        # file.
        # See ibis.udf.vectorized.UserDefinedFunction
        return op.func(*args)

    return scope


@pre_execute.register(ops.AnalyticVectorizedUDF)
@pre_execute.register(ops.AnalyticVectorizedUDF, BaseBackend)
@pre_execute.register(ops.ReductionVectorizedUDF)
@pre_execute.register(ops.ReductionVectorizedUDF, BaseBackend)
def pre_execute_analytic_and_reduction_udf(op, *clients, scope=None, **kwargs):
    input_type = op.input_type
    nargs = len(input_type)

    # An execution rule to handle analytic and reduction UDFs over
    # 1) an ungrouped window,
    # 2) an ungrouped Aggregate node, or
    # 3) an ungrouped custom aggregation context
    @execute_node.register(type(op), *(itertools.repeat(pd.Series, nargs)))
    def execute_udaf_node_no_groupby(op, *args, aggcontext, **kwargs):
        func = op.func
        return aggcontext.agg(args[0], func, *args[1:])

    # An execution rule to handle analytic and reduction UDFs over
    # 1) a grouped window,
    # 2) a grouped Aggregate node, or
    # 3) a grouped custom aggregation context
    @execute_node.register(type(op), *(itertools.repeat(SeriesGroupBy, nargs)))
    def execute_udaf_node_groupby(op, *args, aggcontext, **kwargs):
        func = op.func
        if isinstance(aggcontext, Transform):
            # We are aggregating over an unbounded (and GROUPED) window,
            # which uses a Transform aggregation context.
            # We need to do some pre-processing to func and args so that
            # Transform can pull data out of the SeriesGroupBys in args.

            # Construct a generator that yields the next group of data
            # for every argument excluding the first (pandas performs
            # the iteration for the first argument) for each argument
            # that is a SeriesGroupBy.
            iters = create_gens_from_args_groupby(args[1:])

            # TODO: Unify calling convension here to be more like
            # window
            def aggregator(first, *rest):
                # map(next, *rest) gets the inputs for the next group
                # TODO: might be inefficient to do this on every call
                return func(first, *map(next, rest))

            return aggcontext.agg(args[0], aggregator, *iters)
        else:
            # We are either:
            # 1) Aggregating over a bounded window, which uses a Window
            #  aggregation context
            # 2) Aggregating over a custom aggregation context
            # 3) Aggregating using an Aggregate node (with GROUPING), which
            #   uses a Summarize aggregation context
            # No pre-processing to be done for any case.
            return aggcontext.agg(args[0], func, *args[1:])

    return scope



import pandas.testing as tm
import pyspark.sql.functions as F
import pytest
from pyspark.sql.window import Window

import ibis


@pytest.mark.parametrize(
    ('ibis_windows', 'spark_range'),
    [
        ([(ibis.interval(hours=1), 0)], (-3600, 0)),  # 1h back looking window
        ([(ibis.interval(hours=2), 0)], (-7200, 0)),  # 2h back looking window
        (
            [(0, ibis.interval(hours=1))],
            (0, 3600),
        ),  # 1h forward looking window
    ],
    indirect=['ibis_windows'],
)
def test_time_indexed_window(client, ibis_windows, spark_range):
    table = client.table('time_indexed_table')
    result = table.mutate(
        mean=table['value'].mean().over(ibis_windows[0])
    ).compile()
    result_pd = result.toPandas()
    spark_table = table.compile()
    spark_window = (
        Window.partitionBy('key')
        .orderBy(F.col('time').cast('long'))
        .rangeBetween(*spark_range)
    )
    expected = spark_table.withColumn(
        'mean',
        F.mean(spark_table['value']).over(spark_window),
    ).toPandas()
    tm.assert_frame_equal(result_pd, expected)


@pytest.mark.parametrize(
    ('ibis_windows', 'spark_range'),
    [
        (
            [(ibis.interval(hours=1), 0), (ibis.interval(hours=2), 0)],
            [(-3600, 0), (-7200, 0)],
        ),
    ],
    indirect=['ibis_windows'],
)
def test_multiple_windows(client, ibis_windows, spark_range):
    table = client.table('time_indexed_table')
    result = table.mutate(
        mean_1h=table['value'].mean().over(ibis_windows[0]),
        mean_2h=table['value'].mean().over(ibis_windows[1]),
    ).compile()
    result_pd = result.toPandas()

    spark_table = table.compile()
    spark_window = (
        Window.partitionBy('key')
        .orderBy(F.col('time').cast('long'))
        .rangeBetween(*spark_range[0])
    )
    spark_window_2 = (
        Window.partitionBy('key')
        .orderBy(F.col('time').cast('long'))
        .rangeBetween(*spark_range[1])
    )
    expected = (
        spark_table.withColumn(
            'mean_1h',
            F.mean(spark_table['value']).over(spark_window),
        )
        .withColumn(
            'mean_2h',
            F.mean(spark_table['value']).over(spark_window_2),
        )
        .toPandas()
    )
    tm.assert_frame_equal(result_pd, expected)



import pytest
from multipledispatch import Dispatcher
from multipledispatch.conflict import AmbiguityWarning

from ibis.backends.pandas.dispatcher import TwoLevelDispatcher


class A1:
    pass


class A2(A1):
    pass


class A3(A2):
    pass


class B1:
    pass


class B2(B1):
    pass


class B3(B2):
    pass


@pytest.fixture
def foo_dispatchers():

    foo = TwoLevelDispatcher('foo', doc='Test dispatcher foo')
    foo_m = Dispatcher('foo_m', doc='Control dispatcher foo_m')

    @foo.register(A1, B1)
    @foo_m.register(A1, B1)
    def foo0(x, y):
        return 0

    @foo.register(A1, B2)
    @foo_m.register(A1, B2)
    def foo1(x, y):
        return 1

    @foo.register(A2, B1)
    @foo_m.register(A2, B1)
    def foo2(x, y):
        return 2

    @foo.register(A2, B2)
    @foo_m.register(A2, B2)
    def foo3(x, y):
        return 3

    @foo.register(
        (A1, A2),
    )
    @foo_m.register(
        (A1, A2),
    )
    def foo4(x):
        return 4

    return foo, foo_m


@pytest.fixture
def foo(foo_dispatchers):
    return foo_dispatchers[0]


@pytest.fixture
def foo_m(foo_dispatchers):
    return foo_dispatchers[1]


def test_cache(foo, mocker):
    """Test that cache is properly set after calling with args."""

    spy = mocker.spy(foo, 'dispatch')
    a1, b1 = A1(), B1()

    assert (A1, B1) not in foo._cache
    foo(a1, b1)
    assert (A1, B1) in foo._cache
    foo(a1, b1)
    spy.assert_called_once_with(A1, B1)


def test_dispatch(foo, mocker):
    """Test that calling dispatcher with a signature that is registered
    does not trigger a linear search through dispatch_iter."""

    spy = mocker.spy(foo, 'dispatch_iter')

    # This should not trigger a linear search
    foo(A1(), B1())
    assert not spy.called, (
        "Calling dispatcher with registered signature should "
        "not trigger linear search"
    )

    foo(A3(), B3())
    spy.assert_called_once_with(A3, B3)


@pytest.mark.parametrize(
    'args',
    [
        (A1(), B1()),
        (A1(), B2()),
        (A1(), B3()),
        (A2(), B1()),
        (A2(), B2()),
        (A2(), B3()),
        (A3(), B1()),
        (A3(), B2()),
        (A3(), B3()),
        (A1(),),
        (A2(),),
        (A3(),),
    ],
)
def test_registered(foo_dispatchers, args):
    foo, foo_m = foo_dispatchers
    assert foo(*args) == foo_m(*args)


def test_ordering(foo, foo_m):
    assert foo.ordering == foo_m.ordering


def test_funcs(foo, foo_m):
    assert foo.funcs == foo_m.funcs


@pytest.mark.parametrize(
    'args', [(B1(),), (B2(),), (A1(), A1()), (A1(), A2(), A3())]
)
def test_unregistered(foo, args):
    with pytest.raises(
        NotImplementedError, match="Could not find signature for foo.*"
    ):
        foo(*args)


def test_ambiguities_warning():
    bar = TwoLevelDispatcher('bar')

    bar.register(A1, B1)(lambda a, b: 0)
    bar.register(A1, B2)(lambda a, b: 1)
    bar.register(A2, B1)(lambda a, b: 2)

    with pytest.warns(AmbiguityWarning, match=".*Consider.*\n\n.*(A2, B2).*"):
        bar.reorder()


def test_ambiguities_no_warning():
    bar = TwoLevelDispatcher('bar')

    bar.register(A1, B1)(lambda a, b: 0)
    bar.register(A1, B2)(lambda a, b: 1)
    bar.register(A2, B1)(lambda a, b: 2)
    bar.register(A2, B2)(lambda a, b: 3)

    with pytest.warns(None) as warnings:
        bar.reorder()

    assert len(warnings) == 0



import pandas as pd
import pandas.testing as tm
import pyspark.sql.functions as F
import pytest
from pyspark.sql import Window

import ibis


@pytest.mark.parametrize(
    ('ibis_windows', 'spark_range'),
    [
        ([(ibis.interval(hours=1), 0)], (-3600, 0)),  # 1h back looking window
        ([(ibis.interval(hours=2), 0)], (-7200, 0)),  # 2h back looking window
        (
            [(0, ibis.interval(hours=1))],
            (0, 3600),
        ),  # 1h forward looking window
        (
            [(ibis.interval(hours=1), ibis.interval(hours=1))],
            (-3600, 3600),
        ),  # both forward and trailing
    ],
    indirect=['ibis_windows'],
)
def test_window_with_timecontext(client, ibis_windows, spark_range):
    """Test context adjustment for trailing / range window

    We expand context according to window sizes, for example, for a table of:
    time       value
    2020-01-01   a
    2020-01-02   b
    2020-01-03   c
    2020-01-04   d
    with context = (2020-01-03, 2002-01-04) trailing count for 1 day will be:
    time       value  count
    2020-01-03   c      2
    2020-01-04   d      2
    trailing count for 2 days will be:
    time       value  count
    2020-01-03   c      3
    2020-01-04   d      3
    with context = (2020-01-01, 2002-01-02) count for 1 day forward looking
    window will be:
    time       value  count
    2020-01-01   a      2
    2020-01-02   b      2
    """
    table = client.table('time_indexed_table')
    context = (
        pd.Timestamp('20170102 07:00:00', tz='UTC'),
        pd.Timestamp('20170103', tz='UTC'),
    )
    result_pd = table.mutate(
        count=table['value'].count().over(ibis_windows[0])
    ).execute(timecontext=context)
    spark_table = table.compile()
    spark_window = (
        Window.partitionBy('key')
        .orderBy(F.col('time').cast('long'))
        .rangeBetween(*spark_range)
    )
    expected = spark_table.withColumn(
        'count',
        F.count(spark_table['value']).over(spark_window),
    ).toPandas()
    expected = expected[
        expected.time.between(*(t.tz_convert(None) for t in context))
    ].reset_index(drop=True)
    tm.assert_frame_equal(result_pd, expected)


@pytest.mark.parametrize(
    ('ibis_windows', 'spark_range'),
    [([(None, 0)], (Window.unboundedPreceding, 0))],
    indirect=['ibis_windows'],
)
def test_cumulative_window(client, ibis_windows, spark_range):
    """Test context adjustment for cumulative window

    For cumulative window, by defination we should look back infinately.
    When data is trimmed by time context, we define the limit of looking
    back is the start time of given time context. Thus for a table of
    time       value
    2020-01-01   a
    2020-01-02   b
    2020-01-03   c
    2020-01-04   d
    with context = (2020-01-02, 2002-01-03) cumulative count will be:
    time       value  count
    2020-01-02   b      1
    2020-01-03   c      2
    """
    table = client.table('time_indexed_table')
    context = (
        pd.Timestamp('20170102 07:00:00', tz='UTC'),
        pd.Timestamp('20170105', tz='UTC'),
    )
    result_pd = table.mutate(
        count_cum=table['value'].count().over(ibis_windows[0])
    ).execute(timecontext=context)

    spark_table = table.compile(timecontext=context)
    spark_window = (
        Window.partitionBy('key')
        .orderBy(F.col('time').cast('long'))
        .rangeBetween(*spark_range)
    )
    expected = spark_table.withColumn(
        'count_cum',
        F.count(spark_table['value']).over(spark_window),
    ).toPandas()
    expected = expected[
        expected.time.between(*(t.tz_convert(None) for t in context))
    ].reset_index(drop=True)
    tm.assert_frame_equal(result_pd, expected)


@pytest.mark.parametrize(
    ('ibis_windows', 'spark_range'),
    [
        (
            [(ibis.interval(hours=1), 0), (ibis.interval(hours=2), 0)],
            [(-3600, 0), (-7200, 0)],
        )
    ],
    indirect=['ibis_windows'],
)
def test_multiple_trailing_window(client, ibis_windows, spark_range):
    """Test context adjustment for multiple trailing window

    When there are multiple window ops, we need to verify contexts are
    adjusted correctly for all windows. In this tests we are constucting
    one trailing window for 1h and another trailng window for 2h
    """
    table = client.table('time_indexed_table')
    context = (
        pd.Timestamp('20170102 07:00:00', tz='UTC'),
        pd.Timestamp('20170105', tz='UTC'),
    )
    result_pd = table.mutate(
        count_1h=table['value'].count().over(ibis_windows[0]),
        count_2h=table['value'].count().over(ibis_windows[1]),
    ).execute(timecontext=context)

    spark_table = table.compile()
    spark_window_1h = (
        Window.partitionBy('key')
        .orderBy(F.col('time').cast('long'))
        .rangeBetween(*spark_range[0])
    )
    spark_window_2h = (
        Window.partitionBy('key')
        .orderBy(F.col('time').cast('long'))
        .rangeBetween(*spark_range[1])
    )
    expected = (
        spark_table.withColumn(
            'count_1h', F.count(spark_table['value']).over(spark_window_1h)
        )
        .withColumn(
            'count_2h', F.count(spark_table['value']).over(spark_window_2h)
        )
        .toPandas()
    )
    expected = expected[
        expected.time.between(*(t.tz_convert(None) for t in context))
    ].reset_index(drop=True)
    tm.assert_frame_equal(result_pd, expected)


@pytest.mark.parametrize(
    ('ibis_windows', 'spark_range'),
    [
        (
            [(ibis.interval(hours=1), 0), (ibis.interval(hours=2), 0)],
            [(-3600, 0), (-7200, 0)],
        )
    ],
    indirect=['ibis_windows'],
)
def test_chained_trailing_window(client, ibis_windows, spark_range):
    """Test context adjustment for chained windows

    When there are chained window ops, we need to verify contexts are
    adjusted correctly for all windows. In this tests we are constucting
    one trailing window for 1h and trailng window on the new column for
    2h
    """
    table = client.table('time_indexed_table')
    context = (
        pd.Timestamp('20170102 07:00:00', tz='UTC'),
        pd.Timestamp('20170105', tz='UTC'),
    )
    table = table.mutate(
        new_col=table['value'].count().over(ibis_windows[0]),
    )
    table = table.mutate(count=table['new_col'].count().over(ibis_windows[1]))
    result_pd = table.execute(timecontext=context)

    spark_table = table.compile()
    spark_window_1h = (
        Window.partitionBy('key')
        .orderBy(F.col('time').cast('long'))
        .rangeBetween(*spark_range[0])
    )
    spark_window_2h = (
        Window.partitionBy('key')
        .orderBy(F.col('time').cast('long'))
        .rangeBetween(*spark_range[1])
    )
    spark_table = spark_table.withColumn(
        'new_col', F.count(spark_table['value']).over(spark_window_1h)
    )
    spark_table = spark_table.withColumn(
        'count', F.count(spark_table['new_col']).over(spark_window_2h)
    )
    expected = spark_table.toPandas()
    expected = expected[
        expected.time.between(*(t.tz_convert(None) for t in context))
    ].reset_index(drop=True)
    tm.assert_frame_equal(result_pd, expected)


@pytest.mark.xfail(
    reason='Issue #2457 Adjust context properly for mixed rolling window,'
    ' cumulative window and non window ops',
    strict=True,
)
@pytest.mark.parametrize(
    ('ibis_windows', 'spark_range'),
    [
        (
            [(ibis.interval(hours=1), 0), (None, 0)],
            [(-3600, 0), (Window.unboundedPreceding, 0)],
        )
    ],
    indirect=['ibis_windows'],
)
def test_rolling_with_cumulative_window(client, ibis_windows, spark_range):
    """Test context adjustment for rolling window and cumulative window

    cumulative window should calculate only with in user's context,
    while rolling window should calculate on expanded context.
    For a rolling window of 1 day,
    time       value
    2020-01-01   a
    2020-01-02   b
    2020-01-03   c
    2020-01-04   d
    with context = (2020-01-02, 2002-01-03), count will be:
    time       value  roll_count cum_count
    2020-01-02   b      2            1
    2020-01-03   c      2            2
    """
    table = client.table('time_indexed_table')
    context = (
        pd.Timestamp('20170102 07:00:00', tz='UTC'),
        pd.Timestamp('20170105', tz='UTC'),
    )
    result_pd = table.mutate(
        count_1h=table['value'].count().over(ibis_windows[0]),
        count_cum=table['value'].count().over(ibis_windows[1]),
    ).execute(timecontext=context)

    spark_table = table.compile()
    spark_window_1h = (
        Window.partitionBy('key')
        .orderBy(F.col('time').cast('long'))
        .rangeBetween(*spark_range[0])
    )
    spark_window_cum = (
        Window.partitionBy('key')
        .orderBy(F.col('time').cast('long'))
        .rangeBetween(*spark_range[1])
    )
    expected = (
        spark_table.withColumn(
            'count_1h', F.count(spark_table['value']).over(spark_window_1h)
        )
        .withColumn(
            'count_cum', F.count(spark_table['value']).over(spark_window_cum)
        )
        .toPandas()
    )
    expected = expected[
        expected.time.between(*(t.tz_convert(None) for t in context))
    ].reset_index(drop=True)
    tm.assert_frame_equal(result_pd, expected)


@pytest.mark.xfail(
    reason='Issue #2457 Adjust context properly for mixed rolling window,'
    ' cumulative window and non window ops',
    strict=True,
)
@pytest.mark.parametrize(
    ('ibis_windows', 'spark_range'),
    [([(ibis.interval(hours=1), 0)], [(-3600, 0)])],
    indirect=['ibis_windows'],
)
def test_rolling_with_non_window_op(client, ibis_windows, spark_range):
    """Test context adjustment for rolling window and non window ops

    non window ops should calculate only with in user's context,
    while rolling window should calculate on expanded context.
    For a rolling window of 1 day, and a `count` aggregation
    time       value
    2020-01-01   a
    2020-01-02   b
    2020-01-03   c
    2020-01-04   d
    with context = (2020-01-02, 2002-01-04), result will be:
    time       value  roll_count    count
    2020-01-02   b      2            3
    2020-01-03   c      2            3
    2020-01-04   d      2            3
    Because there are 3 rows within user context (01-02, 01-04),
    count should return 3 for every row, rather 4, based on the
    adjusted context (01-01, 01-04).
    """
    table = client.table('time_indexed_table')
    context = (
        pd.Timestamp('20170102 07:00:00', tz='UTC'),
        pd.Timestamp('20170105', tz='UTC'),
    )
    result_pd = table.mutate(
        count_1h=table['value'].count().over(ibis_windows[0]),
        count=table['value'].count(),
    ).execute(timecontext=context)

    spark_table = table.compile()
    spark_window_1h = (
        Window.partitionBy('key')
        .orderBy(F.col('time').cast('long'))
        .rangeBetween(*spark_range[0])
    )
    expected = (
        spark_table.withColumn(
            'count_1h', F.count(spark_table['value']).over(spark_window_1h)
        )
        .withColumn('count', F.count(spark_table['value']))
        .toPandas()
    )
    expected = expected[
        expected.time.between(*(t.tz_convert(None) for t in context))
    ].reset_index(drop=True)
    tm.assert_frame_equal(result_pd, expected)


@pytest.mark.xfail(
    reason='Issue #2453 chain mutate() for window op and'
    'non window op throws error for pyspark backend',
    strict=True,
)
def test_complex_window(client):
    """Test window with different sizes
    mix context adjustment for window op that require context
    adjustment and non window op that doesn't adjust context
    """
    table = client.table('time_indexed_table')
    context = (
        pd.Timestamp('20170102 07:00:00', tz='UTC'),
        pd.Timestamp('20170105', tz='UTC'),
    )
    window = ibis.trailing_window(
        preceding=ibis.interval(hours=1), order_by='time', group_by='key'
    )
    window2 = ibis.trailing_window(
        preceding=ibis.interval(hours=2), order_by='time', group_by='key'
    )
    window_cum = ibis.cumulative_window(order_by='time', group_by='key')
    # context should be adjusted accordingly for each window
    result_pd = (
        table.mutate(
            count_1h=table['value'].count().over(window),
            count_2h=table['value'].count().over(window2),
            count_cum=table['value'].count().over(window_cum),
        )
        .mutate(count=table['value'].count())
        .execute(timecontext=context)
    )

    df = table.execute()
    expected_win_1h = (
        df.set_index('time')
        .groupby('key')
        .value.rolling('1h', closed='both')
        .count()
        .rename('count_1h')
        .astype(int)
    )
    expected_win_2h = (
        df.set_index('time')
        .groupby('key')
        .value.rolling('2h', closed='both')
        .count()
        .rename('count_2h')
        .astype(int)
    )
    expected_cum_win = (
        df.set_index('time')
        .groupby('key')
        .value.expanding()
        .count()
        .rename('count_cum')
        .astype(int)
    )
    df = df.set_index('time')
    df = df.assign(
        count_1h=expected_win_1h.sort_index(level=['time', 'key']).reset_index(
            level='key', drop=True
        )
    )
    df = df.assign(
        count_2h=expected_win_2h.sort_index(level=['time', 'key']).reset_index(
            level='key', drop=True
        )
    )
    df = df.assign(
        count_cum=expected_cum_win.sort_index(
            level=['time', 'key']
        ).reset_index(level='key', drop=True)
    )
    df['count'] = df.groupby(['key'])['value'].transform('count')
    df = df.reset_index()
    expected = (
        df[df.time.between(*(t.tz_convert(None) for t in context))]
        .sort_values(['key'])
        .reset_index(drop=True)
    )
    tm.assert_frame_equal(result_pd, expected)



import pytest

import ibis
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.expr.types as ir
from ibis.tests.util import assert_equal, assert_pickle_roundtrip


def test_ifelse(table):
    bools = table.g.isnull()
    result = bools.ifelse("foo", "bar")
    assert isinstance(result, ir.StringColumn)


@pytest.mark.xfail(raises=AssertionError, reason='NYT')
def test_ifelse_literal():
    assert False


def test_simple_case_expr(table):
    case1, result1 = "foo", table.a
    case2, result2 = "bar", table.c
    default_result = table.b

    expr1 = table.g.lower().cases(
        [(case1, result1), (case2, result2)], default=default_result
    )

    expr2 = (
        table.g.lower()
        .case()
        .when(case1, result1)
        .when(case2, result2)
        .else_(default_result)
        .end()
    )

    assert_equal(expr1, expr2)
    assert isinstance(expr1, ir.IntegerColumn)


def test_multiple_case_expr(table):
    case1 = table.a == 5
    case2 = table.b == 128
    case3 = table.c == 1000

    result1 = table.f
    result2 = table.b * 2
    result3 = table.e

    default = table.d

    expr = (
        ibis.case()
        .when(case1, result1)
        .when(case2, result2)
        .when(case3, result3)
        .else_(default)
        .end()
    )

    op = expr.op()
    assert isinstance(expr, ir.FloatingColumn)
    assert isinstance(op, ops.SearchedCase)
    assert op.default is default


def test_pickle_multiple_case_node(table):
    case1 = table.a == 5
    case2 = table.b == 128
    case3 = table.c == 1000

    result1 = table.f
    result2 = table.b * 2
    result3 = table.e

    default = table.d
    expr = (
        ibis.case()
        .when(case1, result1)
        .when(case2, result2)
        .when(case3, result3)
        .else_(default)
        .end()
    )

    op = expr.op()
    assert_pickle_roundtrip(op)


@pytest.mark.xfail(raises=AssertionError, reason='NYT')
def test_simple_case_no_default():
    # TODO: this conflicts with the null else cases below. Make a decision
    # about what to do, what to make the default behavior based on what the
    # user provides. SQL behavior is to use NULL when nothing else
    # provided. The .replace convenience API could use the field values as
    # the default, getting us around this issue.
    assert False


def test_simple_case_null_else(table):
    expr = table.g.case().when("foo", "bar").end()
    op = expr.op()

    assert isinstance(expr, ir.StringColumn)
    assert isinstance(op.default, ir.ValueExpr)
    assert isinstance(op.default.op(), ops.Cast)
    assert op.default.op().to == dt.string


def test_multiple_case_null_else(table):
    expr = ibis.case().when(table.g == "foo", "bar").end()
    op = expr.op()

    assert isinstance(expr, ir.StringColumn)
    assert isinstance(op.default, ir.ValueExpr)
    assert isinstance(op.default.op(), ops.Cast)
    assert op.default.op().to == dt.string


@pytest.mark.xfail(raises=AssertionError, reason='NYT')
def test_case_type_precedence():
    assert False


@pytest.mark.xfail(raises=AssertionError, reason='NYT')
def test_no_implicit_cast_possible():
    assert False


def test_case_mixed_type():
    t0 = ibis.table(
        [('one', 'string'), ('two', 'double'), ('three', 'int32')],
        name='my_data',
    )

    expr = (
        t0.three.case()
        .when(0, 'low')
        .when(1, 'high')
        .else_('null')
        .end()
        .name('label')
    )
    result = t0[expr]
    assert result['label'].type().equals(dt.string)



import decimal
import functools
import math
import operator
from operator import methodcaller

import dask.dataframe as dd
import numpy as np
import pandas as pd
import pytest
from dask.dataframe.utils import tm  # noqa: E402

import ibis
import ibis.expr.datatypes as dt  # noqa: E402

from ...execution import execute


@pytest.mark.parametrize(
    'op',
    [
        # comparison
        operator.eq,
        operator.ne,
        operator.lt,
        operator.le,
        operator.gt,
        operator.ge,
    ],
)
def test_binary_operations(t, df, op):
    expr = op(t.plain_float64, t.plain_int64)
    result = expr.compile()
    expected = op(df.plain_float64, df.plain_int64)
    tm.assert_series_equal(result.compute(), expected.compute())


@pytest.mark.parametrize('op', [operator.and_, operator.or_, operator.xor])
def test_binary_boolean_operations(t, df, op):
    expr = op(t.plain_int64 == 1, t.plain_int64 == 2)
    result = expr.compile()
    expected = op(df.plain_int64 == 1, df.plain_int64 == 2)
    tm.assert_series_equal(result.compute(), expected.compute())


def operate(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except decimal.InvalidOperation:
            return decimal.Decimal('NaN')

    return wrapper


@pytest.mark.parametrize(
    ('ibis_func', 'dask_func'),
    [
        (methodcaller('round'), lambda x: np.int64(round(x))),
        (
            methodcaller('round', 2),
            lambda x: x.quantize(decimal.Decimal('.00')),
        ),
        (
            methodcaller('round', 0),
            lambda x: x.quantize(decimal.Decimal('0.')),
        ),
        (methodcaller('ceil'), lambda x: decimal.Decimal(math.ceil(x))),
        (methodcaller('floor'), lambda x: decimal.Decimal(math.floor(x))),
        (methodcaller('exp'), methodcaller('exp')),
        (
            methodcaller('sign'),
            lambda x: x if not x else decimal.Decimal(1).copy_sign(x),
        ),
        (methodcaller('sqrt'), operate(lambda x: x.sqrt())),
        (
            methodcaller('log', 2),
            operate(lambda x: x.ln() / decimal.Decimal(2).ln()),
        ),
        (methodcaller('ln'), operate(lambda x: x.ln())),
        (
            methodcaller('log2'),
            operate(lambda x: x.ln() / decimal.Decimal(2).ln()),
        ),
        (methodcaller('log10'), operate(lambda x: x.log10())),
    ],
)
def test_math_functions_decimal(t, df, ibis_func, dask_func):
    dtype = dt.Decimal(12, 3)
    result = ibis_func(t.float64_as_strings.cast(dtype)).compile()
    context = decimal.Context(prec=dtype.precision)
    expected = df.float64_as_strings.apply(
        lambda x: context.create_decimal(x).quantize(
            decimal.Decimal(
                '{}.{}'.format(
                    '0' * (dtype.precision - dtype.scale), '0' * dtype.scale
                )
            )
        ),
        meta=("float64_as_strings", "object"),
    ).apply(dask_func, meta=("float64_as_strings", "object"))
    # dask.dataframe.Series doesn't do direct item assignment
    # TODO - maybe use .where instead
    computed_result = result.compute()
    computed_result[computed_result.apply(math.isnan)] = -99999
    computed_expected = expected.compute()
    computed_expected[computed_expected.apply(math.isnan)] = -99999
    # result[result.apply(math.isnan)] = -99999
    # expected[expected.apply(math.isnan)] = -99999
    tm.assert_series_equal(computed_result, computed_expected)


def test_round_decimal_with_negative_places(t, df):
    type = dt.Decimal(12, 3)
    expr = t.float64_as_strings.cast(type).round(-1)
    result = expr.compile()
    expected = dd.from_pandas(
        pd.Series(
            list(map(decimal.Decimal, ['1.0E+2', '2.3E+2', '-1.00E+3'])),
            name='float64_as_strings',
        ),
        npartitions=1,
    )
    tm.assert_series_equal(result.compute(), expected.compute())


@pytest.mark.xfail(
    raises=NotImplementedError,
    reason='TODO - arrays - #2553'
    # Need an ops.MultiQuantile execution func that dispatches on ndarrays
)
@pytest.mark.parametrize(
    ('ibis_func', 'dask_func'),
    [
        (
            lambda x: x.quantile([0.25, 0.75]),
            lambda x: list(x.quantile([0.25, 0.75])),
        )
    ],
)
@pytest.mark.parametrize('column', ['float64_with_zeros', 'int64_with_zeros'])
def test_quantile_list(t, df, ibis_func, dask_func, column):
    expr = ibis_func(t[column])
    result = expr.compile()
    expected = dask_func(df[column])
    assert result == expected


@pytest.mark.parametrize(
    ('ibis_func', 'dask_func'),
    [
        (lambda x: x.quantile(0), lambda x: x.quantile(0)),
        (lambda x: x.quantile(1), lambda x: x.quantile(1)),
        (
            lambda x: x.quantile(0.5),
            lambda x: x.quantile(0.5),
        ),
    ],
)
def test_quantile_scalar(t, df, ibis_func, dask_func):
    # TODO - interpolation
    result = ibis_func(t.float64_with_zeros).compile()
    expected = dask_func(df.float64_with_zeros)
    assert result.compute() == expected.compute()

    result = ibis_func(t.int64_with_zeros).compile()
    expected = dask_func(df.int64_with_zeros)
    assert result.compute() == expected.compute()


@pytest.mark.parametrize(
    ('ibis_func', 'exc'),
    [
        # no lower/upper specified
        (lambda x: x.clip(), ValueError),
        # out of range on quantile
        (lambda x: x.quantile(5.0), ValueError),
        # invalid interpolation arg
        (lambda x: x.quantile(0.5, interpolation='foo'), ValueError),
    ],
)
def test_arraylike_functions_transform_errors(t, df, ibis_func, exc):
    with pytest.raises(exc):
        ibis_func(t.float64_with_zeros).execute()


@pytest.mark.xfail(
    raises=NotImplementedError,
    reason='TODO - arrays - #2553'
    # Need an ops.MultiQuantile execution func that dispatches on ndarrays
)
def test_quantile_array_access(client, t, df):
    quantile = t.float64_with_zeros.quantile([0.25, 0.5])
    expr = quantile[0], quantile[1]
    result = tuple(map(client.execute, expr))
    expected = tuple(df.float64_with_zeros.quantile([0.25, 0.5]))
    assert result == expected


def test_ifelse_returning_bool():
    one = ibis.literal(1)
    two = ibis.literal(2)
    true = ibis.literal(True)
    false = ibis.literal(False)
    expr = ibis.ifelse(one + one == two, true, false)
    result = execute(expr)
    assert result is True



import contextlib
import warnings

import sqlalchemy
import sqlalchemy.dialects.mysql as mysql

import ibis.expr.datatypes as dt
from ibis.backends.base.sql.alchemy import BaseAlchemyBackend

from .compiler import MySQLCompiler


class Backend(BaseAlchemyBackend):
    name = 'mysql'
    compiler = MySQLCompiler

    def connect(
        self,
        host='localhost',
        user=None,
        password=None,
        port=3306,
        database=None,
        url=None,
        driver='pymysql',
    ):

        """Create an Ibis client located at `user`:`password`@`host`:`port`
        connected to a MySQL database named `database`.

        Parameters
        ----------
        host : string, default 'localhost'
        user : string, default None
        password : string, default None
        port : string or integer, default 3306
        database : string, default None
        url : string, default None
            Complete SQLAlchemy connection string. If passed, the other
            connection arguments are ignored.
        driver : string, default 'pymysql'

        Returns
        -------
        Backend

        Examples
        --------
        >>> import os
        >>> import getpass
        >>> host = os.environ.get('IBIS_TEST_MYSQL_HOST', 'localhost')
        >>> user = os.environ.get('IBIS_TEST_MYSQL_USER', getpass.getuser())
        >>> password = os.environ.get('IBIS_TEST_MYSQL_PASSWORD')
        >>> database = os.environ.get('IBIS_TEST_MYSQL_DATABASE',
        ...                           'ibis_testing')
        >>> con = connect(
        ...     database=database,
        ...     host=host,
        ...     user=user,
        ...     password=password
        ... )
        >>> con.list_tables()  # doctest: +ELLIPSIS
        [...]
        >>> t = con.table('functional_alltypes')
        >>> t
        MySQLTable[table]
          name: functional_alltypes
          schema:
            index : int64
            Unnamed: 0 : int64
            id : int32
            bool_col : int8
            tinyint_col : int8
            smallint_col : int16
            int_col : int32
            bigint_col : int64
            float_col : float32
            double_col : float64
            date_string_col : string
            string_col : string
            timestamp_col : timestamp
            year : int32
            month : int32
        """
        if driver != 'pymysql':
            raise NotImplementedError(
                'pymysql is currently the only supported driver'
            )
        alchemy_url = self._build_alchemy_url(
            url=url,
            host=host,
            port=port,
            user=user,
            password=password,
            database=database,
            driver=f'mysql+{driver}',
        )

        new_backend = super().connect(sqlalchemy.create_engine(alchemy_url))
        new_backend.database_name = alchemy_url.database

        return new_backend

    @contextlib.contextmanager
    def begin(self):
        with super().begin() as bind:
            previous_timezone = bind.execute(
                'SELECT @@session.time_zone'
            ).scalar()
            try:
                bind.execute("SET @@session.time_zone = 'UTC'")
            except Exception as e:
                warnings.warn(f"Couldn't set mysql timezone: {str(e)}")

            try:
                yield bind
            finally:
                query = "SET @@session.time_zone = '{}'"
                bind.execute(query.format(previous_timezone))

    def table(self, name, database=None, schema=None):
        """Create a table expression that references a particular a table
        called `name` in a MySQL database called `database`.

        Parameters
        ----------
        name : str
            The name of the table to retrieve.
        database : str, optional
            The database in which the table referred to by `name` resides. If
            ``None`` then the ``current_database`` is used.
        schema : str, optional
            The schema in which the table resides.  If ``None`` then the
            `public` schema is assumed.

        Returns
        -------
        table : TableExpr
            A table expression.
        """
        if database is not None and database != self.current_database:
            return self.database(name=database).table(name=name, schema=schema)
        else:
            alch_table = self._get_sqla_table(name, schema=schema)
            node = self.table_class(alch_table, self, self._schemas.get(name))
            return self.table_expr_class(node)


# TODO(kszucs): unsigned integers


@dt.dtype.register((mysql.DOUBLE, mysql.REAL))
def mysql_double(satype, nullable=True):
    return dt.Double(nullable=nullable)


@dt.dtype.register(mysql.FLOAT)
def mysql_float(satype, nullable=True):
    return dt.Float(nullable=nullable)


@dt.dtype.register(mysql.TINYINT)
def mysql_tinyint(satype, nullable=True):
    return dt.Int8(nullable=nullable)


@dt.dtype.register(mysql.BLOB)
def mysql_blob(satype, nullable=True):
    return dt.Binary(nullable=nullable)



""" Tests for macaddr and inet data types"""

import pytest
from pytest import param

import ibis

# add here backends that support json types
net_types_supported = ['postgres']


@pytest.mark.parametrize('data', [param({'status': True}, id='status')])
@pytest.mark.only_on_backends(net_types_supported)
def test_macaddr(backend, con, data, alltypes):
    macaddr_value = '00:0a:95:9d:68:16'
    lit = ibis.literal(macaddr_value, type='macaddr').name('tmp')
    expr = alltypes[[alltypes.id, lit]].head(1)
    df = expr.execute()
    assert df['tmp'].iloc[0] == macaddr_value


@pytest.mark.parametrize('data', [param({'status': True}, id='status')])
@pytest.mark.only_on_backends(net_types_supported)
def test_inet(backend, con, data, alltypes):
    inet_value = '00:0a:95:9d:68:16'
    lit = ibis.literal(inet_value, type='inet').name('tmp')
    expr = alltypes[[alltypes.id, lit]].head(1)
    df = expr.execute()
    assert df['tmp'].iloc[0] == inet_value



import numpy as np
import pandas as pd
import pytest
from pandas.util import testing as tm
from pytest import param

from ..aggcontext import Summarize, window_agg_udf

df = pd.DataFrame(
    {
        'id': [1, 2, 1, 2],
        'v1': [1.0, 2.0, 3.0, 4.0],
        'v2': [10.0, 20.0, 30.0, 40.0],
    }
)


@pytest.mark.parametrize(
    ('agg_fn', 'expected_fn'),
    [
        param(
            lambda v1: v1.mean(),
            lambda df: df['v1'].mean(),
            id='udf',
        ),
        param(
            'mean',
            lambda df: df['v1'].mean(),
            id='string',
        ),
    ],
)
def test_summarize_single_series(agg_fn, expected_fn):
    """Test Summarize.agg operating on a single Series."""

    aggcontext = Summarize()

    result = aggcontext.agg(df['v1'], agg_fn)
    expected = expected_fn(df)

    assert result == expected


@pytest.mark.parametrize(
    ('agg_fn', 'expected_fn'),
    [
        param(
            lambda v1: v1.mean(),
            lambda df: df['v1'].mean(),
            id='udf',
        ),
        param(
            'mean',
            lambda df: df['v1'].mean(),
            id='string',
        ),
    ],
)
def test_summarize_single_seriesgroupby(agg_fn, expected_fn):
    """Test Summarize.agg operating on a single SeriesGroupBy."""

    aggcontext = Summarize()

    df_grouped = df.sort_values('id').groupby('id')
    result = aggcontext.agg(df_grouped['v1'], agg_fn)

    expected = expected_fn(df_grouped)

    tm.assert_series_equal(result, expected)


@pytest.mark.parametrize(
    ('agg_fn', 'expected_fn'),
    [
        param(
            lambda v1, v2: v1.mean() - v2.mean(),
            lambda df: df['v1'].mean() - df['v2'].mean(),
            id='two-column',
        ),
        # Two columns, but only the second one is actually used in UDF
        param(
            lambda v1, v2: v2.mean(),
            lambda df: df['v2'].mean(),
            id='redundant-column',
        ),
    ],
)
def test_summarize_multiple_series(agg_fn, expected_fn):
    """Test Summarize.agg operating on many Series."""

    aggcontext = Summarize()

    args = [df['v1'], df['v2']]
    result = aggcontext.agg(args[0], agg_fn, *args[1:])

    expected = expected_fn(df)

    assert result == expected


@pytest.mark.parametrize(
    'param',
    [
        (
            pd.Series([True, True, True, True]),
            pd.Series([1.0, 2.0, 2.0, 3.0]),
        ),
        (
            pd.Series([False, True, True, False]),
            pd.Series([np.NaN, 2.0, 2.0, np.NaN]),
        ),
    ],
)
def test_window_agg_udf(param):
    """Test passing custom window indices for window aggregation."""

    mask, expected = param

    grouped_data = df.sort_values('id').groupby('id')['v1']
    result_index = grouped_data.obj.index

    window_lower_indices = pd.Series([0, 0, 2, 2])
    window_upper_indices = pd.Series([1, 2, 3, 4])

    result = window_agg_udf(
        grouped_data,
        lambda s: s.mean(),
        window_lower_indices,
        window_upper_indices,
        mask,
        result_index,
        dtype='float',
        max_lookback=None,
    )

    expected.index = grouped_data.obj.index

    tm.assert_series_equal(result, expected)


def test_window_agg_udf_different_freq():
    """Test that window_agg_udf works when the window series and data series
    have different frequencies.
    """

    time = pd.Series([pd.Timestamp('20200101'), pd.Timestamp('20200201')])
    data = pd.Series([1, 2, 3, 4, 5, 6])
    window_lower_indices = pd.Series([0, 4])
    window_upper_indices = pd.Series([5, 7])
    mask = pd.Series([True, True])
    result_index = time.index

    result = window_agg_udf(
        data,
        lambda s: s.mean(),
        window_lower_indices,
        window_upper_indices,
        mask,
        result_index,
        'float',
        None,
    )

    expected = pd.Series([data.iloc[0:5].mean(), data.iloc[4:7].mean()])

    tm.assert_series_equal(result, expected)



"""Code for computing window functions with ibis and pandas."""

import operator
import re
from typing import Any, Callable, List, NoReturn, Optional, Union

import pandas as pd
import toolz
from multipledispatch import Dispatcher
from pandas.core.groupby import SeriesGroupBy

import ibis.common.exceptions as com
import ibis.expr.operations as ops
import ibis.expr.window as win
from ibis.expr.scope import Scope
from ibis.expr.timecontext import (
    construct_time_context_aware_series,
    get_time_col,
)
from ibis.expr.typing import TimeContext

from .. import aggcontext as agg_ctx
from ..aggcontext import AggregationContext
from ..core import (
    compute_time_context,
    date_types,
    execute,
    integer_types,
    simple_types,
    timedelta_types,
    timestamp_types,
)
from ..dispatch import execute_node, pre_execute
from ..execution import util


def _post_process_empty(
    result: Any,
    parent: pd.DataFrame,
    order_by: List[str],
    group_by: List[str],
    timecontext: Optional[TimeContext],
) -> pd.Series:
    # This is the post process of the no groupby nor orderby window
    # `result` could be a Series, DataFrame, or a scalar. generated
    # by `agg` method of class `Window`. For window without grouby or
    # orderby, `agg` calls pands method directly. So if timecontext is
    # present, we need to insert 'time' column into index for trimming the
    # result. For cases when grouby or orderby is present, `agg` calls
    # Ibis method `window_agg_built_in` and `window_agg_udf`, time
    # context is already inserted there.
    assert not order_by and not group_by
    if isinstance(result, (pd.Series, pd.DataFrame)):
        if timecontext:
            result = construct_time_context_aware_series(result, parent)
        return result
    else:
        # `result` is a scalar when a reduction operation is being
        # applied over the window, since reduction operations are N->1
        # in this case we do not need to trim result by timecontext,
        # just expand reduction result to be a Series with `index`.
        index = parent.index
        result = pd.Series([result]).repeat(len(index))
        result.index = index
        return result


def _post_process_group_by(
    series: pd.Series,
    parent: pd.DataFrame,
    order_by: List[str],
    group_by: List[str],
    timecontext: Optional[TimeContext],
) -> pd.Series:
    assert not order_by and group_by
    return series


def _post_process_order_by(
    series,
    parent: pd.DataFrame,
    order_by: List[str],
    group_by: List[str],
    timecontext: Optional[TimeContext],
) -> pd.Series:
    assert order_by and not group_by
    indexed_parent = parent.set_index(order_by)
    index = indexed_parent.index
    names = index.names
    if len(names) > 1:
        series = series.reorder_levels(names)
    series = series.iloc[index.argsort(kind='mergesort')]
    return series


def _post_process_group_by_order_by(
    series: pd.Series,
    parent: pd.DataFrame,
    order_by: List[str],
    group_by: List[str],
    timecontext: Optional[TimeContext],
) -> pd.Series:
    indexed_parent = parent.set_index(group_by + order_by, append=True)
    index = indexed_parent.index

    # get the names of the levels that will be in the result
    series_index_names = frozenset(series.index.names)

    # get the levels common to series.index, in the order that they occur in
    # the parent's index
    reordered_levels = [
        name for name in index.names if name in series_index_names
    ]

    if len(reordered_levels) > 1:
        series = series.reorder_levels(reordered_levels)
    return series


get_aggcontext = Dispatcher('get_aggcontext')


@get_aggcontext.register(object)
def get_aggcontext_default(
    window,
    *,
    scope,
    operand,
    parent,
    group_by,
    order_by,
    **kwargs,
) -> NoReturn:
    raise NotImplementedError(
        f"get_aggcontext is not implemented for {type(window).__name__}"
    )


@get_aggcontext.register(win.Window)
def get_aggcontext_window(
    window,
    *,
    scope,
    operand,
    parent,
    group_by,
    order_by,
    **kwargs,
) -> AggregationContext:
    # no order by or group by: default summarization aggcontext
    #
    # if we're reducing and we have an order by expression then we need to
    # expand or roll.
    #
    # otherwise we're transforming
    output_type = operand.type()

    aggcontext: agg_ctx.AggregationContext
    if not group_by and not order_by:
        aggcontext = agg_ctx.Summarize(parent=parent, output_type=output_type)
    elif (
        isinstance(
            operand.op(), (ops.Reduction, ops.CumulativeOp, ops.Any, ops.All)
        )
        and order_by
    ):
        # XXX(phillipc): What a horror show
        preceding = window.preceding
        if preceding is not None:
            max_lookback = window.max_lookback
            assert not isinstance(operand.op(), ops.CumulativeOp)
            aggcontext = agg_ctx.Moving(
                preceding,
                max_lookback,
                parent=parent,
                group_by=group_by,
                order_by=order_by,
                output_type=output_type,
            )
        else:
            # expanding window
            aggcontext = agg_ctx.Cumulative(
                parent=parent,
                group_by=group_by,
                order_by=order_by,
                output_type=output_type,
            )
    else:
        # groupby transform (window with a partition by clause in SQL parlance)
        aggcontext = agg_ctx.Transform(
            parent=parent,
            group_by=group_by,
            order_by=order_by,
            output_type=output_type,
        )

    return aggcontext


def trim_window_result(
    data: Union[pd.Series, pd.DataFrame], timecontext: Optional[TimeContext]
):
    """Trim data within time range defined by timecontext

    This is a util function used in ``execute_window_op``, where time
    context might be adjusted for calculation. Data must be trimmed
    within the original time context before return.
    `data` is a pd.Series with Multiindex for most cases, for multi
    column udf result, `data` could be a pd.DataFrame

    Params
    ------
    data: pd.Series or pd.DataFrame
    timecontext: Optional[TimeContext]

    Returns:
    ------
    a trimmed pd.Series or or pd.DataFrame with the same Multiindex
    as data's

    """
    # noop if timecontext is None
    if not timecontext:
        return data
    assert isinstance(
        data, (pd.Series, pd.DataFrame)
    ), 'window computed columns is not a pd.Series nor a pd.DataFrame'

    # reset multiindex, convert Series into a DataFrame
    df = data.reset_index()

    # Filter the data, here we preserve the time index so that when user is
    # computing a single column, the computation and the relevant time
    # indexes are returned.
    time_col = get_time_col()
    if time_col not in df:
        return data

    subset = df.loc[df[time_col].between(*timecontext)]

    # Get columns to set for index
    if isinstance(data, pd.Series):
        # if Series dosen't contain a name, reset_index will assign
        # '0' as the column name for the column of value
        name = data.name if data.name else 0
        index_columns = list(subset.columns.difference([name]))
    else:
        name = data.columns
        index_columns = list(subset.columns.difference(name))

    # set the correct index for return Series / DataFrame
    indexed_subset = subset.set_index(index_columns)
    return indexed_subset[name]


@execute_node.register(ops.WindowOp, pd.Series, win.Window)
def execute_window_op(
    op,
    data,
    window,
    scope: Scope = None,
    timecontext: Optional[TimeContext] = None,
    aggcontext=None,
    clients=None,
    **kwargs,
):
    operand = op.expr
    # pre execute "manually" here because otherwise we wouldn't pickup
    # relevant scope changes from the child operand since we're managing
    # execution of that by hand
    operand_op = operand.op()

    adjusted_timecontext = None
    if timecontext:
        arg_timecontexts = compute_time_context(
            op, timecontext=timecontext, clients=clients
        )
        # timecontext is the original time context required by parent node
        # of this WindowOp, while adjusted_timecontext is the adjusted context
        # of this Window, since we are doing a manual execution here, use
        # adjusted_timecontext in later execution phases
        adjusted_timecontext = arg_timecontexts[0]

    pre_executed_scope = pre_execute(
        operand_op,
        *clients,
        scope=scope,
        timecontext=adjusted_timecontext,
        aggcontext=aggcontext,
        **kwargs,
    )
    if scope is None:
        scope = pre_executed_scope
    else:
        scope = scope.merge_scope(pre_executed_scope)
    (root,) = op.root_tables()
    root_expr = root.to_expr()

    data = execute(
        root_expr,
        scope=scope,
        timecontext=adjusted_timecontext,
        clients=clients,
        aggcontext=aggcontext,
        **kwargs,
    )
    following = window.following
    order_by = window._order_by

    if (
        order_by
        and following != 0
        and not isinstance(operand_op, ops.ShiftBase)
    ):
        raise com.OperationNotDefinedError(
            'Window functions affected by following with order_by are not '
            'implemented'
        )

    group_by = window._group_by
    grouping_keys = [
        key_op.name
        if isinstance(key_op, ops.TableColumn)
        else execute(
            key,
            scope=scope,
            clients=clients,
            timecontext=adjusted_timecontext,
            aggcontext=aggcontext,
            **kwargs,
        )
        for key, key_op in zip(
            group_by, map(operator.methodcaller('op'), group_by)
        )
    ]

    order_by = window._order_by
    if not order_by:
        ordering_keys = []

    post_process: Callable[
        [Any, pd.DataFrame, List[str], List[str], Optional[TimeContext]],
        pd.Series,
    ]
    if group_by:
        if order_by:
            (
                sorted_df,
                grouping_keys,
                ordering_keys,
            ) = util.compute_sorted_frame(
                data,
                order_by,
                group_by=group_by,
                timecontext=adjusted_timecontext,
                **kwargs,
            )
            source = sorted_df.groupby(grouping_keys, sort=True)
            post_process = _post_process_group_by_order_by
        else:
            source = data.groupby(grouping_keys, sort=False)
            post_process = _post_process_group_by
    else:
        if order_by:
            source, grouping_keys, ordering_keys = util.compute_sorted_frame(
                data, order_by, timecontext=adjusted_timecontext, **kwargs
            )
            post_process = _post_process_order_by
        else:
            source = data
            post_process = _post_process_empty

    # Here groupby object should be add to the corresponding node in scope
    # for execution, data will be overwrite to a groupby object, so we
    # force an update regardless of time context
    new_scope = scope.merge_scopes(
        [
            Scope({t: source}, adjusted_timecontext)
            for t in operand.op().root_tables()
        ],
        overwrite=True,
    )

    aggcontext = get_aggcontext(
        window,
        scope=scope,
        operand=operand,
        parent=source,
        group_by=grouping_keys,
        order_by=ordering_keys,
        **kwargs,
    )
    result = execute(
        operand,
        scope=new_scope,
        timecontext=adjusted_timecontext,
        aggcontext=aggcontext,
        clients=clients,
        **kwargs,
    )
    result = post_process(
        result,
        data,
        ordering_keys,
        grouping_keys,
        adjusted_timecontext,
    )
    assert len(data) == len(
        result
    ), 'input data source and computed column do not have the same length'

    # trim data to original time context
    result = trim_window_result(result, timecontext)
    return result


@execute_node.register(
    (ops.CumulativeSum, ops.CumulativeMax, ops.CumulativeMin),
    (pd.Series, SeriesGroupBy),
)
def execute_series_cumulative_sum_min_max(op, data, **kwargs):
    typename = type(op).__name__
    method_name = (
        re.match(r"^Cumulative([A-Za-z_][A-Za-z0-9_]*)$", typename)
        .group(1)
        .lower()
    )
    method = getattr(data, f"cum{method_name}")
    return method()


@execute_node.register(ops.CumulativeMean, (pd.Series, SeriesGroupBy))
def execute_series_cumulative_mean(op, data, **kwargs):
    # TODO: Doesn't handle the case where we've grouped/sorted by. Handling
    # this here would probably require a refactor.
    return data.expanding().mean()


@execute_node.register(ops.CumulativeOp, (pd.Series, SeriesGroupBy))
def execute_series_cumulative_op(op, data, aggcontext=None, **kwargs):
    assert aggcontext is not None, "aggcontext is none in {} operation".format(
        type(op)
    )
    typename = type(op).__name__
    match = re.match(r'^Cumulative([A-Za-z_][A-Za-z0-9_]*)$', typename)
    if match is None:
        raise ValueError(f'Unknown operation {typename}')

    try:
        (operation_name,) = match.groups()
    except ValueError:
        raise ValueError(
            f'More than one operation name found in {typename} class'
        )

    dtype = op.to_expr().type().to_pandas()
    assert isinstance(aggcontext, agg_ctx.Cumulative), f'Got {type()}'
    result = aggcontext.agg(data, operation_name.lower())

    # all expanding window operations are required to be int64 or float64, so
    # we need to cast back to preserve the type of the operation
    try:
        return result.astype(dtype)
    except TypeError:
        return result


def post_lead_lag(result, default):
    if not pd.isnull(default):
        return result.fillna(default)
    return result


@execute_node.register(
    (ops.Lead, ops.Lag),
    (pd.Series, SeriesGroupBy),
    integer_types + (type(None),),
    simple_types + (type(None),),
)
def execute_series_lead_lag(op, data, offset, default, **kwargs):
    func = toolz.identity if isinstance(op, ops.Lag) else operator.neg
    result = data.shift(func(1 if offset is None else offset))
    return post_lead_lag(result, default)


@execute_node.register(
    (ops.Lead, ops.Lag),
    (pd.Series, SeriesGroupBy),
    timedelta_types,
    date_types + timestamp_types + (str, type(None)),
)
def execute_series_lead_lag_timedelta(
    op, data, offset, default, aggcontext=None, **kwargs
):
    """An implementation of shifting a column relative to another one that is
    in units of time rather than rows.
    """
    # lagging adds time (delayed), leading subtracts time (moved up)
    func = operator.add if isinstance(op, ops.Lag) else operator.sub
    group_by = aggcontext.group_by
    order_by = aggcontext.order_by

    # get the parent object from which `data` originated
    parent = aggcontext.parent

    # get the DataFrame from the parent object, handling the DataFrameGroupBy
    # case
    parent_df = getattr(parent, 'obj', parent)

    # index our parent df by grouping and ordering keys
    indexed_original_df = parent_df.set_index(group_by + order_by)

    # perform the time shift
    adjusted_parent_df = parent_df.assign(
        **{k: func(parent_df[k], offset) for k in order_by}
    )

    # index the parent *after* adjustment
    adjusted_indexed_parent = adjusted_parent_df.set_index(group_by + order_by)

    # get the column we care about
    result = adjusted_indexed_parent[getattr(data, 'obj', data).name]

    # reindex the shifted data by the original frame's index
    result = result.reindex(indexed_original_df.index)

    # add a default if necessary
    return post_lead_lag(result, default)


@execute_node.register(ops.FirstValue, pd.Series)
def execute_series_first_value(op, data, **kwargs):
    return data.values[0]


@execute_node.register(ops.FirstValue, SeriesGroupBy)
def execute_series_group_by_first_value(op, data, aggcontext=None, **kwargs):
    return aggcontext.agg(data, 'first')


@execute_node.register(ops.LastValue, pd.Series)
def execute_series_last_value(op, data, **kwargs):
    return data.values[-1]


@execute_node.register(ops.LastValue, SeriesGroupBy)
def execute_series_group_by_last_value(op, data, aggcontext=None, **kwargs):
    return aggcontext.agg(data, 'last')


@execute_node.register(ops.MinRank, (pd.Series, SeriesGroupBy))
def execute_series_min_rank(op, data, **kwargs):
    # TODO(phillipc): Handle ORDER BY
    return data.rank(method='min', ascending=True).astype('int64') - 1


@execute_node.register(ops.DenseRank, (pd.Series, SeriesGroupBy))
def execute_series_dense_rank(op, data, **kwargs):
    # TODO(phillipc): Handle ORDER BY
    return data.rank(method='dense', ascending=True).astype('int64') - 1


@execute_node.register(ops.PercentRank, (pd.Series, SeriesGroupBy))
def execute_series_percent_rank(op, data, **kwargs):
    # TODO(phillipc): Handle ORDER BY
    return data.rank(method='min', ascending=True, pct=True)



from .arrays import *  # noqa: F401,F403
from .decimal import *  # noqa: F401,F403
from .generic import *  # noqa: F401,F403
from .join import *  # noqa: F401,F403
from .maps import *  # noqa: F401,F403
from .selection import *  # noqa: F401,F403
from .strings import *  # noqa: F401,F403
from .structs import *  # noqa: F401,F403
from .temporal import *  # noqa: F401,F403
from .timecontext import *  # noqa: F401,F403
from .window import *  # noqa: F401,F403



import datetime

import numpy as np
import pandas as pd
from pandas.core.groupby import SeriesGroupBy

import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
from ibis.backends.base import BaseBackend
from ibis.expr.scope import Scope

from ..core import (
    date_types,
    integer_types,
    numeric_types,
    timedelta_types,
    timestamp_types,
)
from ..dispatch import execute_node, pre_execute


@execute_node.register(ops.Strftime, pd.Timestamp, str)
def execute_strftime_timestamp_str(op, data, format_string, **kwargs):
    return data.strftime(format_string)


@execute_node.register(ops.Strftime, pd.Series, str)
def execute_strftime_series_str(op, data, format_string, **kwargs):
    return data.dt.strftime(format_string)


@execute_node.register(ops.ExtractTemporalField, pd.Timestamp)
def execute_extract_timestamp_field_timestamp(op, data, **kwargs):
    field_name = type(op).__name__.lower().replace('extract', '')
    return getattr(data, field_name)


@execute_node.register(ops.ExtractTemporalField, pd.Series)
def execute_extract_timestamp_field_series(op, data, **kwargs):
    field_name = type(op).__name__.lower().replace('extract', '')
    if field_name == 'weekofyear':
        return data.dt.isocalendar().week.astype(np.int32)
    return getattr(data.dt, field_name).astype(np.int32)


@execute_node.register(ops.ExtractMillisecond, pd.Timestamp)
def execute_extract_millisecond_timestamp(op, data, **kwargs):
    return int(data.microsecond // 1000.0)


@execute_node.register(ops.ExtractMillisecond, pd.Series)
def execute_extract_millisecond_series(op, data, **kwargs):
    return (data.dt.microsecond // 1000).astype(np.int32)


@execute_node.register(ops.ExtractEpochSeconds, (pd.Timestamp, pd.Series))
def execute_epoch_seconds(op, data, **kwargs):
    return data.astype('int64') // int(1e9)


@execute_node.register(
    ops.BetweenTime,
    pd.Series,
    (pd.Series, str, datetime.time),
    (pd.Series, str, datetime.time),
)
def execute_between_time(op, data, lower, upper, **kwargs):
    indexer = pd.DatetimeIndex(data).indexer_between_time(lower, upper)
    result = np.zeros(len(data), dtype=np.bool_)
    result[indexer] = True
    return pd.Series(result)


@execute_node.register(ops.Date, pd.Series)
def execute_timestamp_date(op, data, **kwargs):
    return data.dt.floor('d')


@execute_node.register((ops.TimestampTruncate, ops.DateTruncate), pd.Series)
def execute_timestamp_truncate(op, data, **kwargs):
    dtype = f'datetime64[{op.unit}]'
    array = data.values.astype(dtype)
    return pd.Series(array, name=data.name)


OFFSET_CLASS = {
    "Y": pd.offsets.DateOffset,
    "Q": pd.offsets.DateOffset,
    "M": pd.offsets.DateOffset,
    "W": pd.offsets.DateOffset,
    # all other units are timedelta64s
}


@execute_node.register(ops.IntervalFromInteger, pd.Series)
def execute_interval_from_integer_series(op, data, **kwargs):
    unit = op.unit
    resolution = f"{op.resolution}s"
    cls = OFFSET_CLASS.get(unit, None)

    # fast path for timedelta conversion
    if cls is None:
        return data.astype(f"timedelta64[{unit}]")
    return data.apply(
        lambda n, cls=cls, resolution=resolution: cls(**{resolution: n})
    )


@execute_node.register(ops.IntervalFromInteger, integer_types)
def execute_interval_from_integer_integer_types(op, data, **kwargs):
    unit = op.unit
    resolution = f"{op.resolution}s"
    cls = OFFSET_CLASS.get(unit, None)

    if cls is None:
        return pd.Timedelta(data, unit=unit)
    return cls(**{resolution: data})


@execute_node.register(ops.Cast, pd.Series, dt.Interval)
def execute_cast_integer_to_interval_series(op, data, type, **kwargs):
    to = op.to
    unit = to.unit
    resolution = f"{to.resolution}s"
    cls = OFFSET_CLASS.get(unit, None)

    if cls is None:
        return data.astype(f"timedelta64[{unit}]")
    return data.apply(
        lambda n, cls=cls, resolution=resolution: cls(**{resolution: n})
    )


@execute_node.register(ops.Cast, integer_types, dt.Interval)
def execute_cast_integer_to_interval_integer_types(op, data, type, **kwargs):
    to = op.to
    unit = to.unit
    resolution = f"{to.resolution}s"
    cls = OFFSET_CLASS.get(unit, None)

    if cls is None:
        return pd.Timedelta(data, unit=unit)
    return cls(**{resolution: data})


@execute_node.register(ops.TimestampAdd, timestamp_types, timedelta_types)
def execute_timestamp_add_datetime_timedelta(op, left, right, **kwargs):
    return pd.Timestamp(left) + pd.Timedelta(right)


@execute_node.register(ops.TimestampAdd, timestamp_types, pd.Series)
def execute_timestamp_add_datetime_series(op, left, right, **kwargs):
    return pd.Timestamp(left) + right


@execute_node.register(ops.IntervalAdd, timedelta_types, timedelta_types)
def execute_interval_add_delta_delta(op, left, right, **kwargs):
    return op.op(pd.Timedelta(left), pd.Timedelta(right))


@execute_node.register(ops.IntervalAdd, timedelta_types, pd.Series)
@execute_node.register(
    ops.IntervalMultiply, timedelta_types, numeric_types + (pd.Series,)
)
def execute_interval_add_multiply_delta_series(op, left, right, **kwargs):
    return op.op(pd.Timedelta(left), right)


@execute_node.register(
    (ops.TimestampAdd, ops.IntervalAdd), pd.Series, timedelta_types
)
def execute_timestamp_interval_add_series_delta(op, left, right, **kwargs):
    return left + pd.Timedelta(right)


@execute_node.register(
    (ops.TimestampAdd, ops.IntervalAdd), pd.Series, pd.Series
)
def execute_timestamp_interval_add_series_series(op, left, right, **kwargs):
    return left + right


@execute_node.register(ops.TimestampSub, timestamp_types, timedelta_types)
def execute_timestamp_sub_datetime_timedelta(op, left, right, **kwargs):
    return pd.Timestamp(left) - pd.Timedelta(right)


@execute_node.register(
    (ops.TimestampDiff, ops.TimestampSub), timestamp_types, pd.Series
)
def execute_timestamp_diff_sub_datetime_series(op, left, right, **kwargs):
    return pd.Timestamp(left) - right


@execute_node.register(ops.TimestampSub, pd.Series, timedelta_types)
def execute_timestamp_sub_series_timedelta(op, left, right, **kwargs):
    return left - pd.Timedelta(right)


@execute_node.register(
    (ops.TimestampDiff, ops.TimestampSub, ops.IntervalSubtract),
    pd.Series,
    pd.Series,
)
def execute_timestamp_diff_sub_series_series(op, left, right, **kwargs):
    return left - right


@execute_node.register(ops.TimestampDiff, timestamp_types, timestamp_types)
def execute_timestamp_diff_datetime_datetime(op, left, right, **kwargs):
    return pd.Timestamp(left) - pd.Timestamp(right)


@execute_node.register(ops.TimestampDiff, pd.Series, timestamp_types)
def execute_timestamp_diff_series_datetime(op, left, right, **kwargs):
    return left - pd.Timestamp(right)


@execute_node.register(
    ops.IntervalMultiply, pd.Series, numeric_types + (pd.Series,)
)
@execute_node.register(
    ops.IntervalFloorDivide,
    (pd.Timedelta, pd.Series),
    numeric_types + (pd.Series,),
)
def execute_interval_multiply_fdiv_series_numeric(op, left, right, **kwargs):
    return op.op(left, right)


@execute_node.register(ops.TimestampFromUNIX, (pd.Series,) + integer_types)
def execute_timestamp_from_unix(op, data, **kwargs):
    return pd.to_datetime(data, unit=op.unit)


@pre_execute.register(ops.TimestampNow)
@pre_execute.register(ops.TimestampNow, BaseBackend)
def pre_execute_timestamp_now(op, *args, **kwargs):
    timecontext = kwargs.get('timecontext', None)
    return Scope({op: pd.Timestamp('now')}, timecontext)


@execute_node.register(ops.DayOfWeekIndex, (str, datetime.date))
def execute_day_of_week_index_any(op, value, **kwargs):
    return pd.Timestamp(value).dayofweek


@execute_node.register(ops.DayOfWeekIndex, pd.Series)
def execute_day_of_week_index_series(op, data, **kwargs):
    return data.dt.dayofweek.astype(np.int16)


@execute_node.register(ops.DayOfWeekIndex, SeriesGroupBy)
def execute_day_of_week_index_series_group_by(op, data, **kwargs):
    groupings = data.grouper.groupings
    return data.obj.dt.dayofweek.astype(np.int16).groupby(groupings)


def day_name(obj):
    """Backwards compatible name of day getting function.

    Parameters
    ----------
    obj : Union[Series, pd.Timestamp]

    Returns
    -------
    str
        The name of the day corresponding to `obj`
    """
    try:
        return obj.day_name()
    except AttributeError:
        return obj.weekday_name


@execute_node.register(ops.DayOfWeekName, (str, datetime.date))
def execute_day_of_week_name_any(op, value, **kwargs):
    return day_name(pd.Timestamp(value))


@execute_node.register(ops.DayOfWeekName, pd.Series)
def execute_day_of_week_name_series(op, data, **kwargs):
    return day_name(data.dt)


@execute_node.register(ops.DayOfWeekName, SeriesGroupBy)
def execute_day_of_week_name_series_group_by(op, data, **kwargs):
    return day_name(data.obj.dt).groupby(data.grouper.groupings)


@execute_node.register(ops.DateSub, date_types, timedelta_types)
@execute_node.register((ops.DateDiff, ops.DateSub), date_types, pd.Series)
@execute_node.register(ops.DateSub, pd.Series, timedelta_types)
@execute_node.register((ops.DateDiff, ops.DateSub), pd.Series, pd.Series)
@execute_node.register(ops.DateDiff, date_types, date_types)
@execute_node.register(ops.DateDiff, pd.Series, date_types)
def execute_date_sub_diff(op, left, right, **kwargs):
    return left - right


@execute_node.register(ops.DateAdd, pd.Series, timedelta_types)
@execute_node.register(ops.DateAdd, timedelta_types, pd.Series)
@execute_node.register(ops.DateAdd, pd.Series, pd.Series)
@execute_node.register(ops.DateAdd, date_types, timedelta_types)
@execute_node.register(ops.DateAdd, timedelta_types, date_types)
@execute_node.register(ops.DateAdd, date_types, pd.Series)
@execute_node.register(ops.DateAdd, pd.Series, date_types)
def execute_date_add(op, left, right, **kwargs):
    return left + right



import re

import numpy as np
import pandas as pd

import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.expr.types as ir

fully_qualified_re = re.compile(r"(.*)\.(?:`(.*)`|(.*))")
base_typename_re = re.compile(r"(\w+)")


_clickhouse_dtypes = {
    'Null': dt.Null,
    'Nothing': dt.Null,
    'UInt8': dt.UInt8,
    'UInt16': dt.UInt16,
    'UInt32': dt.UInt32,
    'UInt64': dt.UInt64,
    'Int8': dt.Int8,
    'Int16': dt.Int16,
    'Int32': dt.Int32,
    'Int64': dt.Int64,
    'Float32': dt.Float32,
    'Float64': dt.Float64,
    'String': dt.String,
    'FixedString': dt.String,
    'Date': dt.Date,
    'DateTime': dt.Timestamp,
}
_ibis_dtypes = {v: k for k, v in _clickhouse_dtypes.items()}
_ibis_dtypes[dt.String] = 'String'


class ClickhouseDataType:

    __slots__ = 'typename', 'nullable'

    def __init__(self, typename, nullable=False):
        m = base_typename_re.match(typename)
        base_typename = m.groups()[0]
        if base_typename not in _clickhouse_dtypes:
            raise com.UnsupportedBackendType(typename)
        self.typename = base_typename
        self.nullable = nullable

    def __str__(self):
        if self.nullable:
            return f'Nullable({self.typename})'
        else:
            return self.typename

    def __repr__(self):
        return f'<Clickhouse {str(self)}>'

    @classmethod
    def parse(cls, spec):
        # TODO(kszucs): spare parsing, depends on clickhouse-driver#22
        if spec.startswith('Nullable'):
            return cls(spec[9:-1], nullable=True)
        else:
            return cls(spec)

    def to_ibis(self):
        return _clickhouse_dtypes[self.typename](nullable=self.nullable)

    @classmethod
    def from_ibis(cls, dtype, nullable=None):
        typename = _ibis_dtypes[type(dtype)]
        if nullable is None:
            nullable = dtype.nullable
        return cls(typename, nullable=nullable)


@dt.dtype.register(ClickhouseDataType)
def clickhouse_to_ibis_dtype(clickhouse_dtype):
    return clickhouse_dtype.to_ibis()


class ClickhouseTable(ir.TableExpr):
    """References a physical table in Clickhouse"""

    @property
    def _qualified_name(self):
        return self.op().args[0]

    @property
    def _unqualified_name(self):
        return self._match_name()[1]

    @property
    def _client(self):
        return self.op().args[2]

    def _match_name(self):
        m = fully_qualified_re.match(self._qualified_name)
        if not m:
            raise com.IbisError(
                'Cannot determine database name from {}'.format(
                    self._qualified_name
                )
            )
        db, quoted, unquoted = m.groups()
        return db, quoted or unquoted

    @property
    def _database(self):
        return self._match_name()[0]

    def invalidate_metadata(self):
        self._client.invalidate_metadata(self._qualified_name)

    def metadata(self):
        """
        Return parsed results of DESCRIBE FORMATTED statement

        Returns
        -------
        meta : TableMetadata
        """
        return self._client.describe_formatted(self._qualified_name)

    describe_formatted = metadata

    @property
    def name(self):
        return self.op().name

    def insert(self, obj, **kwargs):
        from .identifiers import quote_identifier

        schema = self.schema()

        assert isinstance(obj, pd.DataFrame)
        assert set(schema.names) >= set(obj.columns)

        columns = ', '.join(map(quote_identifier, obj.columns))
        query = 'INSERT INTO {table} ({columns}) VALUES'.format(
            table=self._qualified_name, columns=columns
        )

        # convert data columns with datetime64 pandas dtype to native date
        # because clickhouse-driver 0.0.10 does arithmetic operations on it
        obj = obj.copy()
        for col in obj.select_dtypes(include=[np.datetime64]):
            if isinstance(schema[col], dt.Date):
                obj[col] = obj[col].dt.date

        data = obj.to_dict('records')
        return self._client.con.execute(query, data, **kwargs)



import operator
from typing import Any, Optional, Union

import numpy as np
import pandas as pd
import toolz

import ibis.common.exceptions as com
import ibis.util
from ibis.expr import operations as ops
from ibis.expr import types as ir
from ibis.expr.schema import coerce_to_dataframe
from ibis.expr.scope import Scope

from ..core import execute


def compute_sort_key(key, data, timecontext, scope=None, **kwargs):
    by = key.to_expr()
    try:
        if isinstance(by, str):
            return by, None
        return by.get_name(), None
    except com.ExpressionError:
        if scope is None:
            scope = Scope()
        scope = scope.merge_scopes(
            Scope({t: data}, timecontext) for t in by.op().root_tables()
        )
        new_column = execute(by, scope=scope, **kwargs)
        name = ibis.util.guid()
        new_column.name = name
        return name, new_column


def compute_sorted_frame(
    df, order_by, group_by=(), timecontext=None, **kwargs
):
    computed_sort_keys = []
    sort_keys = list(toolz.concatv(group_by, order_by))
    ascending = [getattr(key.op(), 'ascending', True) for key in sort_keys]
    new_columns = {}

    for i, key in enumerate(map(operator.methodcaller('op'), sort_keys)):
        computed_sort_key, temporary_column = compute_sort_key(
            key, df, timecontext, **kwargs
        )
        computed_sort_keys.append(computed_sort_key)

        if temporary_column is not None:
            new_columns[computed_sort_key] = temporary_column

    result = df.assign(**new_columns)
    result = result.sort_values(
        computed_sort_keys, ascending=ascending, kind='mergesort'
    )
    # TODO: we'll eventually need to return this frame with the temporary
    # columns and drop them in the caller (maybe using post_execute?)
    ngrouping_keys = len(group_by)
    return (
        result,
        computed_sort_keys[:ngrouping_keys],
        computed_sort_keys[ngrouping_keys:],
    )


def coerce_to_output(
    result: Any, expr: ir.Expr, index: Optional[pd.Index] = None
) -> Union[pd.Series, pd.DataFrame]:
    """Cast the result to either a Series or DataFrame.

    This method casts result of an execution to a Series or DataFrame,
    depending on the type of the expression and shape of the result.

    Parameters
    ----------
    result: Any
        The result to cast
    expr: ibis.expr.types.Expr
        The expression associated with the result
    index: pd.Index
        Optional. If passed, scalar results will be broadcasted according
        to the index.

    Returns
    -------
    result: A Series or DataFrame

    Examples
    --------
    For dataframe outputs, see ``ibis.util.coerce_to_dataframe``.

    >>> coerce_to_output(pd.Series(1), expr)
    0    1
    Name: result, dtype: int64
    >>> coerce_to_output(1, expr)
    0    1
    Name: result, dtype: int64
    >>> coerce_to_output(1, expr, [1,2,3])
    1    1
    2    1
    3    1
    Name: result, dtype: int64
    >>> coerce_to_output([1,2,3], expr)
    0    [1, 2, 3]
    Name: result, dtype: object
    """
    result_name = getattr(expr, '_name', None)

    if isinstance(expr, (ir.DestructColumn, ir.StructColumn)):
        return coerce_to_dataframe(result, expr.type())
    elif isinstance(expr, (ir.DestructScalar, ir.StructScalar)):
        # Here there are two cases, if this is groupby aggregate,
        # then the result e a Series of tuple/list, or
        # if this is non grouped aggregate, then the result
        return coerce_to_dataframe(result, expr.type())
    elif isinstance(result, pd.Series):
        return result.rename(result_name)
    elif isinstance(expr.op(), ops.Reduction):
        if index is None:
            # Wrap `result` into a single-element Series.
            return pd.Series([result], name=result_name)
        else:
            # Broadcast `result` to a multi-element Series according to the
            # given `index`.
            return pd.Series(
                np.repeat(result, len(index)),
                index=index,
                name=result_name,
            )
    elif isinstance(result, np.ndarray):
        return pd.Series(result, name=result_name)
    else:
        raise ValueError(f"Cannot coerce_to_output. Result: {result}")



import os
from pathlib import Path

import pytest

import ibis
import ibis.expr.types as ir
from ibis.backends.tests.base import BackendTest, RoundAwayFromZero


class TestConf(BackendTest, RoundAwayFromZero):
    supports_arrays = False
    supports_arrays_outside_of_select = supports_arrays
    supports_window_operations = True
    check_dtype = False
    returned_timestamp_unit = 's'

    @staticmethod
    def connect(data_directory: Path):
        path = Path(
            os.environ.get(
                'IBIS_TEST_SQLITE_DATABASE', data_directory / 'ibis_testing.db'
            )
        )
        return ibis.sqlite.connect(str(path))  # type: ignore

    @property
    def functional_alltypes(self) -> ir.TableExpr:
        t = super().functional_alltypes
        return t.mutate(timestamp_col=t.timestamp_col.cast('timestamp'))


@pytest.fixture
def dbpath(data_directory):
    default = str(data_directory / 'ibis_testing.db')
    return os.environ.get('IBIS_TEST_SQLITE_DATABASE', default)


@pytest.fixture
def con(dbpath):
    return ibis.sqlite.connect(dbpath)


@pytest.fixture
def db(con):
    return con.database()


@pytest.fixture
def dialect():
    import sqlalchemy as sa

    return sa.dialects.sqlite.dialect()


@pytest.fixture
def translate(dialect):
    from ibis.backends.sqlite import Backend

    context = Backend.compiler.make_context()
    return lambda expr: str(
        Backend.compiler.translator_class(expr, context)
        .get_result()
        .compile(dialect=dialect, compile_kwargs={'literal_binds': True})
    )


@pytest.fixture
def sqla_compile(dialect):
    return lambda expr: str(
        expr.compile(dialect=dialect, compile_kwargs={'literal_binds': True})
    )


@pytest.fixture
def alltypes(db):
    return db.functional_alltypes


@pytest.fixture
def alltypes_sqla(alltypes):
    return alltypes.op().sqla_table


@pytest.fixture
def df(alltypes):
    return alltypes.execute()



import numpy as np
import pytest

import ibis
import ibis.expr.types as ir


@pytest.mark.xfail_unsupported
@pytest.mark.skip_missing_feature(
    ['supports_arrays', 'supports_arrays_outside_of_select']
)
def test_array_column(backend, alltypes, df):
    expr = ibis.array([alltypes['double_col'], alltypes['double_col']])
    assert isinstance(expr, ir.ArrayColumn)

    result = expr.execute()
    expected = df.apply(
        lambda row: np.array(
            [row['double_col'], row['double_col']], dtype=object
        ),
        axis=1,
    )
    backend.assert_series_equal(result, expected, check_names=False)


@pytest.mark.xfail_unsupported
@pytest.mark.skip_missing_feature(
    ['supports_arrays', 'supports_arrays_outside_of_select']
)
def test_array_scalar(backend, con, alltypes, df):
    expr = ibis.array([1.0, 2.0, 3.0])
    assert isinstance(expr, ir.ArrayScalar)

    result = con.execute(expr)
    expected = np.array([1.0, 2.0, 3.0])

    # This does not check whether `result` is an np.array or a list,
    # because it varies across backends and backend configurations
    assert np.array_equal(result, expected)


@pytest.mark.xfail_unsupported
@pytest.mark.skip_missing_feature(
    ['supports_arrays', 'supports_arrays_outside_of_select']
)
# Issues #2370
@pytest.mark.xfail_backends(['bigquery'])
def test_array_concat(backend, con):
    left = ibis.literal([1, 2, 3])
    right = ibis.literal([2, 1])
    expr = left + right
    result = con.execute(expr)
    expected = np.array([1, 2, 3, 2, 1])

    # This does not check whether `result` is an np.array or a list,
    # because it varies across backends and backend configurations
    assert np.array_equal(result, expected)


@pytest.mark.xfail_unsupported
@pytest.mark.skip_missing_feature(
    ['supports_arrays', 'supports_arrays_outside_of_select']
)
def test_array_length(backend, con):
    expr = ibis.literal([1, 2, 3]).length()
    assert con.execute(expr) == 3


@pytest.mark.xfail_unsupported
@pytest.mark.skip_missing_feature(
    ['supports_arrays', 'supports_arrays_outside_of_select']
)
def test_list_literal(backend, con):
    arr = [1, 2, 3]
    expr = ibis.literal(arr)
    result = con.execute(expr)

    # This does not check whether `result` is an np.array or a list,
    # because it varies across backends and backend configurations
    assert np.array_equal(result, arr)


@pytest.mark.xfail_unsupported
@pytest.mark.skip_missing_feature(
    ['supports_arrays', 'supports_arrays_outside_of_select']
)
def test_np_array_literal(backend, con):
    arr = np.array([1, 2, 3])
    expr = ibis.literal(arr)
    result = con.execute(expr)

    # This does not check whether `result` is an np.array or a list,
    # because it varies across backends and backend configurations
    assert np.array_equal(result, arr)



import ibis
from ibis.expr import datatypes
from ibis.expr.operations import Literal
from ibis.tests.util import assert_pickle_roundtrip


def test_literal_equality_basic():
    a = ibis.literal(1).op()
    b = ibis.literal(1).op()

    assert a == b
    assert hash(a) == hash(b)


def test_literal_equality_int_float():
    # Note: This is different from the Python behavior for int/float comparison
    a = ibis.literal(1).op()
    b = ibis.literal(1.0).op()

    assert a != b


def test_literal_equality_int16_int32():
    # Note: This is different from the Python behavior for int/float comparison
    a = Literal(1, datatypes.int16)
    b = Literal(1, datatypes.int32)

    assert a != b


def test_literal_equality_int_interval():
    a = ibis.literal(1).op()
    b = ibis.interval(seconds=1).op()

    assert a != b


def test_literal_equality_interval():
    a = ibis.interval(seconds=1).op()
    b = ibis.interval(minutes=1).op()

    assert a != b

    # Currently these does't equal, but perhaps should be?
    c = ibis.interval(seconds=60).op()
    d = ibis.interval(minutes=1).op()

    assert c != d


def test_pickle_literal():
    a = Literal(1, datatypes.int16)
    b = Literal(1, datatypes.int32)

    assert_pickle_roundtrip(a)
    assert_pickle_roundtrip(b)


def test_pickle_literal_interval():
    a = ibis.interval(seconds=1).op()

    assert_pickle_roundtrip(a)



import itertools

import dask.dataframe as dd
import dask.dataframe.groupby as ddgb
import numpy as np

import ibis.expr.operations as ops
from ibis.backends.pandas.execution.arrays import (
    execute_array_index,
    execute_array_length,
)

from ..dispatch import execute_node
from .util import TypeRegistrationDict, register_types_to_dispatcher

DASK_DISPATCH_TYPES: TypeRegistrationDict = {
    ops.ArrayLength: [((dd.Series,), execute_array_length)],
    ops.ArrayIndex: [((dd.Series, int), execute_array_index)],
}

register_types_to_dispatcher(execute_node, DASK_DISPATCH_TYPES)


collect_list = dd.Aggregation(
    name="collect_list",
    chunk=lambda s: s.apply(list),
    agg=lambda s0: s0.apply(
        lambda chunks: list(itertools.chain.from_iterable(chunks))
    ),
)


@execute_node.register(ops.ArrayColumn, list)
def execute_array_column(op, cols, **kwargs):
    df = dd.concat(cols, axis=1)
    return df.apply(
        lambda row: np.array(row, dtype=object), axis=1, meta=(None, 'object')
    )


# TODO - aggregations - #2553
@execute_node.register(ops.ArrayCollect, dd.Series)
def execute_array_collect(op, data, aggcontext=None, **kwargs):
    return aggcontext.agg(data, collect_list)


@execute_node.register(ops.ArrayCollect, ddgb.SeriesGroupBy)
def execute_array_collect_grouped_series(op, data, aggcontext=None, **kwargs):
    return data.agg(collect_list)



from typing import Optional

import sqlalchemy as sa
from sqlalchemy.dialects import mysql, postgresql, sqlite
from sqlalchemy.dialects.mysql.base import MySQLDialect
from sqlalchemy.dialects.postgresql.base import PGDialect
from sqlalchemy.dialects.sqlite.base import SQLiteDialect
from sqlalchemy.engine.interfaces import Dialect

import ibis.expr.datatypes as dt
import ibis.expr.schema as sch

from .geospatial import geospatial_supported

if geospatial_supported:
    import geoalchemy2 as ga


def table_from_schema(name, meta, schema, database: Optional[str] = None):
    # Convert Ibis schema to SQLA table
    columns = []

    for colname, dtype in zip(schema.names, schema.types):
        satype = to_sqla_type(dtype)
        column = sa.Column(colname, satype, nullable=dtype.nullable)
        columns.append(column)

    return sa.Table(name, meta, schema=database, *columns)


# TODO(cleanup)
ibis_type_to_sqla = {
    dt.Null: sa.types.NullType,
    dt.Date: sa.Date,
    dt.Time: sa.Time,
    dt.Boolean: sa.Boolean,
    dt.Binary: sa.LargeBinary,
    dt.String: sa.Text,
    dt.Decimal: sa.NUMERIC,
    # Mantissa-based
    dt.Float: sa.Float(precision=24),
    dt.Double: sa.Float(precision=53),
    dt.Int8: sa.SmallInteger,
    dt.Int16: sa.SmallInteger,
    dt.Int32: sa.Integer,
    dt.Int64: sa.BigInteger,
}


def to_sqla_type(itype, type_map=None):
    if type_map is None:
        type_map = ibis_type_to_sqla
    if isinstance(itype, dt.Decimal):
        return sa.types.NUMERIC(itype.precision, itype.scale)
    elif isinstance(itype, dt.Date):
        return sa.Date()
    elif isinstance(itype, dt.Timestamp):
        # SQLAlchemy DateTimes do not store the timezone, just whether the db
        # supports timezones.
        return sa.TIMESTAMP(bool(itype.timezone))
    elif isinstance(itype, dt.Array):
        ibis_type = itype.value_type
        if not isinstance(ibis_type, (dt.Primitive, dt.String)):
            raise TypeError(
                'Type {} is not a primitive type or string type'.format(
                    ibis_type
                )
            )
        return sa.ARRAY(to_sqla_type(ibis_type, type_map=type_map))
    elif geospatial_supported and isinstance(itype, dt.GeoSpatial):
        if itype.geotype == 'geometry':
            return ga.Geometry
        elif itype.geotype == 'geography':
            return ga.Geography
        else:
            return ga.types._GISType
    else:
        return type_map[type(itype)]


@dt.dtype.register(Dialect, sa.types.NullType)
def sa_null(_, satype, nullable=True):
    return dt.null


@dt.dtype.register(Dialect, sa.types.Boolean)
def sa_boolean(_, satype, nullable=True):
    return dt.Boolean(nullable=nullable)


@dt.dtype.register(MySQLDialect, mysql.NUMERIC)
def sa_mysql_numeric(_, satype, nullable=True):
    # https://dev.mysql.com/doc/refman/8.0/en/fixed-point-types.html
    return dt.Decimal(
        satype.precision or 10, satype.scale or 0, nullable=nullable
    )


@dt.dtype.register(PGDialect, postgresql.NUMERIC)
def sa_postgres_numeric(_, satype, nullable=True):
    # PostgreSQL allows any precision for numeric values if not specified,
    # up to the implementation limit. Here, default to the maximum value that
    # can be specified by the user. The scale defaults to zero.
    # https://www.postgresql.org/docs/10/datatype-numeric.html
    return dt.Decimal(
        satype.precision or 1000, satype.scale or 0, nullable=nullable
    )


@dt.dtype.register(Dialect, sa.types.Numeric)
@dt.dtype.register(SQLiteDialect, sqlite.NUMERIC)
def sa_numeric(_, satype, nullable=True):
    return dt.Decimal(satype.precision, satype.scale, nullable=nullable)


@dt.dtype.register(Dialect, sa.types.SmallInteger)
def sa_smallint(_, satype, nullable=True):
    return dt.Int16(nullable=nullable)


@dt.dtype.register(Dialect, sa.types.Integer)
def sa_integer(_, satype, nullable=True):
    return dt.Int32(nullable=nullable)


@dt.dtype.register(Dialect, mysql.TINYINT)
def sa_mysql_tinyint(_, satype, nullable=True):
    return dt.Int8(nullable=nullable)


@dt.dtype.register(Dialect, sa.types.BigInteger)
def sa_bigint(_, satype, nullable=True):
    return dt.Int64(nullable=nullable)


@dt.dtype.register(Dialect, sa.types.Float)
def sa_float(_, satype, nullable=True):
    return dt.Float(nullable=nullable)


@dt.dtype.register(SQLiteDialect, sa.types.Float)
@dt.dtype.register(PGDialect, postgresql.DOUBLE_PRECISION)
def sa_double(_, satype, nullable=True):
    return dt.Double(nullable=nullable)


@dt.dtype.register(PGDialect, postgresql.UUID)
def sa_uuid(_, satype, nullable=True):
    return dt.UUID(nullable=nullable)


@dt.dtype.register(PGDialect, postgresql.MACADDR)
def sa_macaddr(_, satype, nullable=True):
    return dt.MACADDR(nullable=nullable)


@dt.dtype.register(PGDialect, postgresql.INET)
def sa_inet(_, satype, nullable=True):
    return dt.INET(nullable=nullable)


@dt.dtype.register(PGDialect, postgresql.JSON)
def sa_json(_, satype, nullable=True):
    return dt.JSON(nullable=nullable)


@dt.dtype.register(PGDialect, postgresql.JSONB)
def sa_jsonb(_, satype, nullable=True):
    return dt.JSONB(nullable=nullable)


if geospatial_supported:

    @dt.dtype.register(Dialect, (ga.Geometry, ga.types._GISType))
    def ga_geometry(_, gatype, nullable=True):
        t = gatype.geometry_type
        if t == 'POINT':
            return dt.Point(nullable=nullable)
        if t == 'LINESTRING':
            return dt.LineString(nullable=nullable)
        if t == 'POLYGON':
            return dt.Polygon(nullable=nullable)
        if t == 'MULTILINESTRING':
            return dt.MultiLineString(nullable=nullable)
        if t == 'MULTIPOINT':
            return dt.MultiPoint(nullable=nullable)
        if t == 'MULTIPOLYGON':
            return dt.MultiPolygon(nullable=nullable)
        if t == 'GEOMETRY':
            return dt.Geometry(nullable=nullable)
        else:
            raise ValueError(f"Unrecognized geometry type: {t}")


POSTGRES_FIELD_TO_IBIS_UNIT = {
    "YEAR": "Y",
    "MONTH": "M",
    "DAY": "D",
    "HOUR": "h",
    "MINUTE": "m",
    "SECOND": "s",
    "YEAR TO MONTH": "M",
    "DAY TO HOUR": "h",
    "DAY TO MINUTE": "m",
    "DAY TO SECOND": "s",
    "HOUR TO MINUTE": "m",
    "HOUR TO SECOND": "s",
    "MINUTE TO SECOND": "s",
}


@dt.dtype.register(PGDialect, postgresql.INTERVAL)
def sa_postgres_interval(_, satype, nullable=True):
    field = satype.fields.upper()
    unit = POSTGRES_FIELD_TO_IBIS_UNIT.get(field, None)
    if unit is None:
        raise ValueError(f"Unknown PostgreSQL interval field {field!r}")
    elif unit in {"Y", "M"}:
        raise ValueError(
            "Variable length timedeltas are not yet supported with PostgreSQL"
        )
    return dt.Interval(unit=unit, nullable=nullable)


@dt.dtype.register(MySQLDialect, mysql.DOUBLE)
def sa_mysql_double(_, satype, nullable=True):
    # TODO: handle asdecimal=True
    return dt.Double(nullable=nullable)


@dt.dtype.register(Dialect, sa.types.String)
def sa_string(_, satype, nullable=True):
    return dt.String(nullable=nullable)


@dt.dtype.register(Dialect, sa.LargeBinary)
def sa_binary(_, satype, nullable=True):
    return dt.Binary(nullable=nullable)


@dt.dtype.register(Dialect, sa.Time)
def sa_time(_, satype, nullable=True):
    return dt.Time(nullable=nullable)


@dt.dtype.register(Dialect, sa.Date)
def sa_date(_, satype, nullable=True):
    return dt.Date(nullable=nullable)


@dt.dtype.register(Dialect, sa.DateTime)
def sa_datetime(_, satype, nullable=True, default_timezone='UTC'):
    timezone = default_timezone if satype.timezone else None
    return dt.Timestamp(timezone=timezone, nullable=nullable)


@dt.dtype.register(Dialect, sa.ARRAY)
def sa_array(dialect, satype, nullable=True):
    dimensions = satype.dimensions
    if dimensions is not None and dimensions != 1:
        raise NotImplementedError('Nested array types not yet supported')

    value_dtype = dt.dtype(dialect, satype.item_type)
    return dt.Array(value_dtype, nullable=nullable)


@sch.infer.register(sa.Table)
def schema_from_table(table, schema=None):
    """Retrieve an ibis schema from a SQLAlchemy ``Table``.

    Parameters
    ----------
    table : sa.Table

    Returns
    -------
    schema : ibis.expr.datatypes.Schema
        An ibis schema corresponding to the types of the columns in `table`.
    """
    schema = schema if schema is not None else {}
    pairs = []
    for name, column in zip(table.columns.keys(), table.columns):
        if name in schema:
            dtype = dt.dtype(schema[name])
        else:
            dtype = dt.dtype(
                getattr(table.bind, 'dialect', Dialect()),
                column.type,
                nullable=column.nullable,
            )
        pairs.append((name, dtype))
    return sch.schema(pairs)



import ibis.expr.operations as ops
import ibis.expr.schema as sch
from ibis.backends.base import Database


class AlchemyDatabase(Database):
    """

    Attributes
    ----------
    client : AlchemyClient

    """

    def table(self, name, schema=None):
        return self.client.table(name, schema=schema)


class AlchemyTable(ops.DatabaseTable):
    def __init__(self, table, source, schema=None):
        schema = sch.infer(table, schema=schema)
        super().__init__(table.name, schema, source)
        self.sqla_table = table



from .base import DDL, DML
from .query_builder import (
    Compiler,
    Select,
    SelectBuilder,
    TableSetFormatter,
    Union,
)
from .translator import ExprTranslator, QueryContext

__all__ = (
    'Compiler',
    'Select',
    'SelectBuilder',
    'Union',
    'TableSetFormatter',
    'ExprTranslator',
    'QueryContext',
    'DML',
    'DDL',
)
